import base64
import json
import re
import shutil
import tempfile
from collections.abc import Mapping
from pathlib import Path
from typing import Any

from einops import rearrange
from jax import numpy as jnp
from jaxtyping import Array, Float
from tokenizers import Tokenizer

from lalamo.common import ParameterPath
from lalamo.modules import (
    Attention,
    DenseMLP,
    FullPrecisionLinear,
    Identity,
    LinearBase,
    MLPBase,
    Normalization,
    Transformer,
    TransformerLayer,
)
from lalamo.modules.audio.fishaudio import DescriptAudioCodec, FishAudioTextDecoder
from lalamo.modules.audio.fishaudio.fishaudio_common import (
    FishAudioSpecialInferenceTokens,
)
from lalamo.modules.audio.fishaudio.fishaudio_consts import (
    FISH_TIKTOKEN_PATTERN,
    IM_END_TOKEN,
)
from lalamo.modules.audio.fishaudio.fishaudio_modules import (
    ConvNeXtBlock,
    DACDecoder,
    DACDecoderBlock,
    DownsampleResidualVectorQuantize,
    ResidualUnit,
    ResidualVectorQuantize,
    Upsampler,
    UpsamplingBlock,
    VectorQuantize,
)

from .common import load_parameters
from .huggingface import load_rmsnorm, load_tied_embedding
from .nanocodec_loaders import (
    load_causal_conv1d,
    load_causal_transpose_conv1d,
    load_snake1d,
)
from .torch_utils import fuse_weight_norm_conv1d_as_linear


def _permute_for_rope_rotate_half(
    weight: Array,
    num_heads: int,
    head_dim: int,
) -> Array:
    """Permute weight matrix from interleaved RoPE format to rotate-half format.

    This converts weights trained with the interleaved RoPE format:
        interleaved: [-x1, x0, -x3, x2, -x5, x4, ...]
    to the rotate-half format used by standard RoPE:
        rotate-half: [-x_{d/2}, -x_{d/2+1}, ..., x_0, x_1, ...]

    The transformation reorders the output dimensions of Q/K projections so that
    the first half of each head's dimensions and the second half are grouped together,
    rather than being interleaved.

    Code is inspired by similar transformation from:
    https://github.com/huggingface/transformers/blob/e42587f596181396e1c4b63660abf0c736b10dae/src/transformers/models/llama/convert_llama_weights_to_hf.py
    """
    if len(weight.shape) == 1:
        # For 1D vectors: swap interleaved pairs to grouped halves
        return rearrange(weight, "(half_dim pair) -> (pair half_dim)", pair=2)

    out_features, _ = weight.shape
    assert out_features == num_heads * head_dim, (
        f"Output features {out_features} must equal num_heads * head_dim = {num_heads * head_dim}"
    )
    # For 2D matrices: swap interleaved pairs to grouped halves within each head
    return rearrange(
        weight,
        "(heads half_dim pair) in_features -> (heads pair half_dim) in_features",
        heads=num_heads,
        pair=2,
    )


def _permute_qkv_for_rope_rotate_half(
    qkv_weight: Float[Array, "q_dim+k_dim+v_dim in_features"],
    num_heads: int,
    num_groups: int,
    head_dim: int,
) -> Array:
    """Permute fused QKV weight matrix from interleaved RoPE to rotate-half format.

    For grouped query attention (GQA), Q has num_heads while K/V have num_groups.
    Only Q and K need permutation (they use RoPE). V is unchanged.

    Args:
        qkv_weight: Fused QKV weight of shape (q_dim + k_dim + v_dim, in_features)
                    where q_dim = num_heads * head_dim, k_dim = v_dim = num_groups * head_dim
        num_heads: Number of query heads.
        num_groups: Number of key/value heads (groups for GQA).
        head_dim: Dimension per head.

    Returns:
        Permuted QKV weight with Q and K converted to rotate-half format.
    """
    q_dim = num_heads * head_dim
    k_dim = num_groups * head_dim

    q_weight = qkv_weight[:q_dim, :]
    k_weight = qkv_weight[q_dim : q_dim + k_dim, :]
    v_weight = qkv_weight[q_dim + k_dim :, :]

    q_weight = _permute_for_rope_rotate_half(q_weight, num_heads, head_dim)
    k_weight = _permute_for_rope_rotate_half(k_weight, num_groups, head_dim)

    return jnp.concatenate([q_weight, k_weight, v_weight], axis=0)


def _fuse_full_precision_weights(
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    sublayers_to_fuse: list[str] | None,
) -> Array:
    if sublayers_to_fuse is None:
        return weights_dict[path / "weight"]

    weights = [weights_dict[path / layer_name / "weight"] for layer_name in sublayers_to_fuse]
    return jnp.concatenate(weights, axis=0)


def load_linear_and_fuse_scaling(
    module: LinearBase,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
    sublayers_to_fuse: list[str] | None = None,
    scaling_to_fuse: Array | None = None,
) -> LinearBase:
    """Load linear layer directly or fuse several sum-matrices into one linear layer.
    Additionally fuse final result with scaling weights that would follow after the layer.
    Args:
        module: target linear module into which weights will be loaded
        weights_dict: mapping with weights
        path: path to linear layer within the given weights mapping
        sublayers_to_fuse: optional list of names of matrices that we want to fuse into single linear layer
        scaling_to_fuze: optional array of scales we want to fuse into given linear module

    Returns:
        Linear layer with weights loaded into it
    """
    assert isinstance(module, FullPrecisionLinear)
    if not module.has_biases:
        if sublayers_to_fuse:
            paths_to_check = [path / proj / "bias" for proj in sublayers_to_fuse]
        else:
            paths_to_check = path / "bias"
        for p in paths_to_check:
            if p in weights_dict:
                raise ValueError(f"Bias tensor found at {p} but module does not support it.")
        bias = None
    elif sublayers_to_fuse is None:
        bias = weights_dict[path / "bias"]
    else:
        bias = jnp.concatenate(
            [weights_dict[path / proj_name / "bias"] for proj_name in sublayers_to_fuse],
            axis=0,
        )

    weights = _fuse_full_precision_weights(weights_dict, path, sublayers_to_fuse)

    if scaling_to_fuse is not None:
        weights = weights * scaling_to_fuse[:, None]
        if bias is not None:
            bias = bias * scaling_to_fuse

    return load_parameters(lambda m: (m.weights, m.biases), module, (weights, bias))


def load_transformer_block(
    module: Transformer,
    weights_dict: Mapping[str, Array],
    fast: bool = False,
    path: ParameterPath | None = None,
) -> Transformer:
    def load_attention_local(
        attn_module: Attention,
        weights_dict: Mapping[str, Array],
        path: ParameterPath,
        scaling_to_fuze: Array | None = None,
    ) -> Attention:
        qkv_projection = load_linear_and_fuse_scaling(
            attn_module.qkv_projection,
            weights_dict,
            path / "wqkv",
            sublayers_to_fuse=None,
        )
        assert isinstance(qkv_projection, FullPrecisionLinear)

        # Permute QKV weights from interleaved RoPE format to rotate-half format
        permuted_qkv_weights = _permute_qkv_for_rope_rotate_half(
            qkv_projection.weights,
            num_heads=attn_module.num_heads,
            num_groups=attn_module.num_groups,
            head_dim=attn_module.head_dim,
        )
        qkv_projection = load_parameters(
            lambda m: (m.weights,),
            qkv_projection,
            (permuted_qkv_weights,),
        )
        assert isinstance(qkv_projection, FullPrecisionLinear)

        out_projection = load_linear_and_fuse_scaling(
            attn_module.out_projection,
            weights_dict,
            path / "wo",
            scaling_to_fuse=scaling_to_fuze,
        )

        if attn_module.query_norm is not None:
            query_norm = load_rmsnorm(attn_module.query_norm, weights_dict, path / "q_norm")
            permuted_scales = _permute_for_rope_rotate_half(
                query_norm.scales,
                1,
                query_norm.scales.shape[0],
            )
            query_norm = load_parameters(
                lambda m: (m.scales,),
                query_norm,
                (permuted_scales,),
            )
        else:
            query_norm = None

        if attn_module.key_norm is not None:
            key_norm = load_rmsnorm(attn_module.key_norm, weights_dict, path / "k_norm")
            permuted_scales = _permute_for_rope_rotate_half(
                key_norm.scales,
                1,
                key_norm.scales.shape[0],
            )
            key_norm = load_parameters(
                lambda m: (m.scales,),
                key_norm,
                (permuted_scales,),
            )
        else:
            key_norm = None

        return load_parameters(
            lambda m: (m.qkv_projection, m.out_projection, m.query_norm, m.key_norm),
            attn_module,
            (qkv_projection, out_projection, query_norm, key_norm),
        )

    def load_mlp(
        module: MLPBase,
        weights_dict: Mapping[str, Array],
        path: ParameterPath,
        up_proj_key: str,
        gate_proj_key: str,
        down_proj_key: str,
        scaling_to_fuze: Array | None = None,
    ) -> MLPBase:
        assert isinstance(module, DenseMLP)
        # Standard dense MLP with separate sublayers.
        up_projection = load_linear_and_fuse_scaling(
            module.up_projection,
            weights_dict,
            path,
            sublayers_to_fuse=[up_proj_key, gate_proj_key],
        )
        down_projection = load_linear_and_fuse_scaling(
            module.down_projection,
            weights_dict,
            path / down_proj_key,
            scaling_to_fuse=scaling_to_fuze,
        )
        return load_parameters(
            lambda m: (m.up_projection, m.down_projection),
            module,
            (up_projection, down_projection),
        )

    def load_transformer_layer_local(
        module: TransformerLayer,
        weights_dict: Mapping[str, Array],
        path: ParameterPath,
    ) -> TransformerLayer:
        if module.pre_mixer_norm is not None:
            assert isinstance(module.pre_mixer_norm, Normalization)
            pre_mixer_norm = load_rmsnorm(
                module.pre_mixer_norm,
                weights_dict,
                path / "attention_norm",
            )
        else:
            pre_mixer_norm = None

        layer_scale_path = path / "attention_layer_scale" / "gamma"
        layer_scale_weights = weights_dict.get(layer_scale_path, None)
        post_mixer_norm = None

        assert isinstance(module.mixer, Attention)
        attention = load_attention_local(
            attn_module=module.mixer,
            weights_dict=weights_dict,
            path=path / "attention",
            scaling_to_fuze=layer_scale_weights,
        )

        assert isinstance(module.pre_mlp_norm, Normalization)
        pre_mlp_norm = load_rmsnorm(
            module.pre_mlp_norm,
            weights_dict,
            path / "ffn_norm",
        )

        layer_scale_path = path / "ffn_layer_scale" / "gamma"
        layer_scale_weights = weights_dict.get(layer_scale_path, None)
        post_mlp_norm = None

        mlp = load_mlp(
            module=module.mlp,
            weights_dict=weights_dict,
            path=path / "feed_forward",
            up_proj_key="w3",
            gate_proj_key="w1",
            down_proj_key="w2",
            scaling_to_fuze=layer_scale_weights,
        )

        return load_parameters(
            lambda m: (
                m.pre_mixer_norm,
                m.mixer,
                m.post_mixer_norm,
                m.pre_mlp_norm,
                m.mlp,
                m.post_mlp_norm,
            ),
            module,
            (
                pre_mixer_norm,
                attention,
                post_mixer_norm,
                pre_mlp_norm,
                mlp,
                post_mlp_norm,
            ),
        )

    base_path = ParameterPath() if path is None else path

    layers_name = "layers" if not fast else "fast_layers"
    norm_name = "norm" if not fast else "fast_norm"

    transformer_layers = tuple(
        load_transformer_layer_local(layer, weights_dict, base_path / layers_name / i)
        for i, layer in enumerate(module.layers)
    )
    output_norm = load_rmsnorm(module.output_norm, weights_dict, base_path / norm_name)

    module = load_parameters(
        lambda m: (
            m.layers,
            m.output_norm,
        ),
        module,
        (
            transformer_layers,
            output_norm,
        ),
    )

    return module


def load_fish_audio_text_decoding_modules(
    transformer: Transformer,
    output: FullPrecisionLinear,
    weights_dict: Mapping[str, Array],
    fast: bool = False,
) -> tuple[Transformer, FullPrecisionLinear]:
    transformer = load_transformer_block(transformer, weights_dict=weights_dict, fast=fast)

    base_path = ParameterPath()
    output_linear_name = "output" if not fast else "fast_output"
    output_linear = load_linear_and_fuse_scaling(output, weights_dict, base_path / output_linear_name)
    output = load_parameters(
        lambda m: (m,),
        output,
        (output_linear,),
    )

    return (transformer, output)


def load_vector_quantize(
    module: VectorQuantize,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> VectorQuantize:
    """Loads a VectorQuantize module from weights.

    The in_proj and out_proj layers use weight normalization in the original PyTorch model.
    This function fuses the weight_g and weight_v parameters using PyTorch's remove_weight_norm.

    Expected weight structure at path:
        - in_proj.weight_g, in_proj.weight_v, in_proj.bias (unused in decode-only mode)
        - out_proj.weight_g, out_proj.weight_v, out_proj.bias
        - codebook.weight

    Args:
        module: The VectorQuantize module to load weights into.
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Base path for this module's weights.

    Returns:
        VectorQuantize module with loaded weights.
    """
    # Load codebook weights
    codebook_weight = weights_dict[path / "codebook" / "weight"]
    codebook = load_parameters(
        lambda m: (m.weights,),
        module.codebook,
        (codebook_weight,),
    )

    # Load out_proj with weight norm fusion
    # The original is a Conv1d with kernel_size=1, so weight shape is (out, in, 1)
    # Our FullPrecisionLinear expects (out, in), so we remove the kernel dimension
    out_proj_weight, out_proj_bias = fuse_weight_norm_conv1d_as_linear(weights_dict, path / "out_proj")
    # Remove kernel dimension: (out_channels, in_channels, 1) -> (out_channels, in_channels)
    out_proj_weight = rearrange(out_proj_weight, "out_ch in_ch 1 -> out_ch in_ch")
    out_proj = load_parameters(
        lambda m: (m.weights, m.biases),
        module.out_proj,
        (out_proj_weight, out_proj_bias),
    )

    return load_parameters(
        lambda m: (m.codebook, m.out_proj),
        module,
        (codebook, out_proj),
    )


def load_residual_vector_quantize(
    module: ResidualVectorQuantize,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> ResidualVectorQuantize:
    """Loads a ResidualVectorQuantize module from weights.

    Expected weight structure at path:
        - quantizers.0.in_proj.weight_g, quantizers.0.in_proj.weight_v, etc.
        - quantizers.0.out_proj.weight_g, quantizers.0.out_proj.weight_v, etc.
        - quantizers.0.codebook.weight
        - quantizers.1..., quantizers.2..., etc.

    Args:
        module: The ResidualVectorQuantize module to load weights into.
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Base path for this module's weights.

    Returns:
        ResidualVectorQuantize module with loaded weights.
    """
    quantizers = tuple(
        load_vector_quantize(quantizer, weights_dict, path / "quantizers" / i)
        for i, quantizer in enumerate(module.quantizers)
    )

    return load_parameters(
        lambda m: (m.quantizers,),
        module,
        (quantizers,),
    )


def load_convnext_block(
    module: ConvNeXtBlock,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> ConvNeXtBlock:
    """Loads a ConvNeXtBlock module from weights.

    Expected weight structure at path:
        - gamma
        - dwconv.conv.weight
        - dwconv.conv.bias
        - norm.weight
        - norm.bias
        - pwconv1.weight
        - pwconv1.bias
        - pwconv2.weight
        - pwconv2.bias

    Args:
        module: The ConvNeXtBlock module to load weights into.
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Base path for this module's weights.

    Returns:
        ConvNeXtBlock module with loaded weights.
    """

    # Load depthwise conv
    # PyTorch conv weights are (out_channels, in_channels/groups, kernel_size)
    dwconv_weight = weights_dict[path / "dwconv" / "conv" / "weight"]
    dwconv_bias = weights_dict[path / "dwconv" / "conv" / "bias"]
    depthwise_conv = load_parameters(
        lambda m: (m.weights, m.biases),
        module.depthwise_conv,
        (dwconv_weight, dwconv_bias),
    )

    # Load norm (LayerNorm with weight and bias)
    norm_weight = weights_dict[path / "norm" / "weight"]
    norm_bias = weights_dict[path / "norm" / "bias"]
    norm = load_parameters(
        lambda m: (m.scales, m.biases),
        module.norm,
        (norm_weight, norm_bias),
    )

    # Load pointwise conv 1 (Linear layer)
    # PyTorch Linear weight is (out_features, in_features)
    pwconv1_weight = weights_dict[path / "pwconv1" / "weight"]
    pwconv1_bias = weights_dict[path / "pwconv1" / "bias"]
    pointwise_conv_step1 = load_parameters(
        lambda m: (m.weights, m.biases),
        module.pointwise_conv_step1,
        (pwconv1_weight, pwconv1_bias),
    )

    # Load pointwise conv 2 (Linear layer), fusing layer scaling if present
    pwconv2_weight = weights_dict[path / "pwconv2" / "weight"]
    pwconv2_bias = weights_dict[path / "pwconv2" / "bias"]
    layer_scale_path = path / "gamma"
    if layer_scale_path in weights_dict:
        layer_scale = weights_dict[layer_scale_path]
        pwconv2_weight = pwconv2_weight * layer_scale[:, None]
        pwconv2_bias = pwconv2_bias * layer_scale
    pointwise_conv_step2 = load_parameters(
        lambda m: (m.weights, m.biases),
        module.pointwise_conv_step2,
        (pwconv2_weight, pwconv2_bias),
    )

    return load_parameters(
        lambda m: (m.depthwise_conv, m.norm, m.pointwise_conv_step1, m.pointwise_conv_step2),
        module,
        (depthwise_conv, norm, pointwise_conv_step1, pointwise_conv_step2),
    )


def load_upsampling_block(
    module: UpsamplingBlock,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> UpsamplingBlock:
    """Loads an UpsamplingBlock module from weights.

    Expected weight structure at path:
        - 0.conv.weight (transpose conv)
        - 0.conv.bias
        - 1.gamma (ConvNeXt LayerScale)
        - 1.dwconv.conv.weight
        - 1.dwconv.conv.bias
        - 1.norm.weight
        - 1.norm.bias
        - 1.pwconv1.weight
        - 1.pwconv1.bias
        - 1.pwconv2.weight
        - 1.pwconv2.bias

    Args:
        module: The UpsamplingBlock module to load weights into.
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Base path for this module's weights.

    Returns:
        UpsamplingBlock module with loaded weights.
    """
    # Load transpose conv (at index 0)
    trans_conv = load_causal_transpose_conv1d(module.trans_conv, weights_dict, path / "0" / "conv")

    # Load ConvNeXt block (at index 1)
    convnext = load_convnext_block(module.convnext, weights_dict, path / "1")

    return load_parameters(
        lambda m: (m.trans_conv, m.convnext),
        module,
        (trans_conv, convnext),
    )


def load_upsampler(
    module: Upsampler,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> Upsampler:
    """Loads an Upsampler module from weights.

    Expected weight structure at path:
        - 0.0.conv.weight, 0.1.gamma, etc. (first UpsamplingBlock)
        - 1.0.conv.weight, 1.1.gamma, etc. (second UpsamplingBlock)
        - ...

    Args:
        module: The Upsampler module to load weights into.
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Base path for this module's weights.

    Returns:
        Upsampler module with loaded weights.
    """
    blocks = tuple(load_upsampling_block(block, weights_dict, path / i) for i, block in enumerate(module.blocks))

    return load_parameters(
        lambda m: (m.blocks,),
        module,
        (blocks,),
    )


def load_downsample_rvq(
    module: DownsampleResidualVectorQuantize,
    weights_dict: Mapping[str, Array],
    path: ParameterPath | None = None,
) -> DownsampleResidualVectorQuantize:
    """Loads a DownsampleResidualVectorQuantize module from weights.

    This function handles the complete loading of the audio decoder module, including:
    - semantic_quantizer: ResidualVectorQuantize with weight-normalized projections
    - quantizer: ResidualVectorQuantize with weight-normalized projections
    - upsample: Upsampler with transpose convolutions and ConvNeXt blocks
    - post_module: Transformer for post-processing

    Weight normalization is fused using PyTorch's remove_weight_norm internally.

    Expected weight structure:
        semantic_quantizer:
            - semantic_quantizer.quantizers.0.in_proj.weight_g/weight_v/bias
            - semantic_quantizer.quantizers.0.out_proj.weight_g/weight_v/bias
            - semantic_quantizer.quantizers.0.codebook.weight
            - ... (for each quantizer)

        quantizer:
            - quantizer.quantizers.0.in_proj.weight_g/weight_v/bias
            - quantizer.quantizers.0.out_proj.weight_g/weight_v/bias
            - quantizer.quantizers.0.codebook.weight
            - ... (for each quantizer)

        upsample (UpsamplingBlocks):
            - upsample.0.0.conv.weight (transpose conv)
            - upsample.0.0.conv.bias
            - upsample.0.1.gamma (ConvNeXt LayerScale)
            - upsample.0.1.dwconv.conv.weight
            - upsample.0.1.dwconv.conv.bias
            - upsample.0.1.norm.weight
            - upsample.0.1.norm.bias
            - upsample.0.1.pwconv1.weight
            - upsample.0.1.pwconv1.bias
            - upsample.0.1.pwconv2.weight
            - upsample.0.1.pwconv2.bias
            - ... (for each upsampling block)

        post_module (Transformer):
            - [SKIP] post_module.freqs_cis (generated internally, not loaded)
            - [SKIP] post_module.causal_mask (generated internally, not loaded)
            - post_module.layers.0.attention.wqkv.weight
            - post_module.layers.0.attention.wo.weight
            - post_module.layers.0.feed_forward.w1.weight
            - post_module.layers.0.feed_forward.w3.weight
            - post_module.layers.0.feed_forward.w2.weight
            - post_module.layers.0.ffn_norm.weight
            - post_module.layers.0.attention_norm.weight
            - post_module.layers.0.attention_layer_scale.gamma
            - post_module.layers.0.ffn_layer_scale.gamma
            - ... (for each layer)
            - post_module.norm.weight

    Args:
        module: The DownsampleResidualVectorQuantize module to load weights into.
        weights_dict: Dictionary mapping parameter paths to weight arrays.

    Returns:
        DownsampleResidualVectorQuantize module with loaded weights.
    """
    base_path = ParameterPath() if path is None else path

    semantic_quantizer = load_residual_vector_quantize(
        module.semantic_quantizer,
        weights_dict,
        base_path / "semantic_quantizer",
    )

    quantizer = load_residual_vector_quantize(
        module.quantizer,
        weights_dict,
        base_path / "quantizer",
    )

    upsampler = load_upsampler(
        module.upsampler,
        weights_dict,
        base_path / "upsample",
    )

    post_module = load_transformer_block(module.post_module, weights_dict, fast=False, path=base_path / "post_module")

    return load_parameters(
        lambda m: (m.semantic_quantizer, m.quantizer, m.upsampler, m.post_module),
        module,
        (semantic_quantizer, quantizer, upsampler, post_module),
    )


def load_residual_unit(
    module: ResidualUnit,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> ResidualUnit:
    """Loads a ResidualUnit module from weights.

    Expected weight structure at path (PyTorch block is nn.Sequential):
        - block.0.alpha (Snake1d)
        - block.1.conv.parametrizations.weight.original0/1, block.1.conv.bias (CausalWNConv1d, kernel=7)
        - block.2.alpha (Snake1d)
        - block.3.conv.parametrizations.weight.original0/1, block.3.conv.bias (CausalWNConv1d, kernel=1)

    Args:
        module: The ResidualUnit module to load weights into.
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Base path for this module's weights.

    Returns:
        ResidualUnit module with loaded weights.
    """
    snake1 = load_snake1d(module.snake1, weights_dict, path / "block" / "0")
    conv1 = load_causal_conv1d(module.conv1, weights_dict, path / "block" / "1" / "conv")
    snake2 = load_snake1d(module.snake2, weights_dict, path / "block" / "2")
    conv2 = load_causal_conv1d(module.conv2, weights_dict, path / "block" / "3" / "conv")

    return load_parameters(
        lambda m: (m.snake1, m.conv1, m.snake2, m.conv2),
        module,
        (snake1, conv1, snake2, conv2),
    )


def load_audio_decoder_block(
    module: DACDecoderBlock,
    weights_dict: Mapping[str, Array],
    path: ParameterPath,
) -> DACDecoderBlock:
    """Loads an AudioDecoderBlock module from weights.

    Expected weight structure at path (PyTorch block is nn.Sequential):
        - block.0.alpha (Snake1d)
        - block.1.conv.parametrizations.weight.original0/1, block.1.conv.bias (CausalWNConvTranspose1d)
        - block.2.block.0..3 (ResidualUnit, dilation=1)
        - block.3.block.0..3 (ResidualUnit, dilation=3)
        - block.4.block.0..3 (ResidualUnit, dilation=9)

    Args:
        module: The AudioDecoderBlock module to load weights into.
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Base path for this module's weights.

    Returns:
        AudioDecoderBlock module with loaded weights.
    """
    snake = load_snake1d(module.snake, weights_dict, path / "block" / "0")
    trans_conv = load_causal_transpose_conv1d(module.trans_conv, weights_dict, path / "block" / "1" / "conv")
    res_unit1 = load_residual_unit(module.res_unit1, weights_dict, path / "block" / "2")
    res_unit2 = load_residual_unit(module.res_unit2, weights_dict, path / "block" / "3")
    res_unit3 = load_residual_unit(module.res_unit3, weights_dict, path / "block" / "4")

    return load_parameters(
        lambda m: (m.snake, m.trans_conv, m.res_unit1, m.res_unit2, m.res_unit3),
        module,
        (snake, trans_conv, res_unit1, res_unit2, res_unit3),
    )


def load_audio_decoder(
    module: DACDecoder,
    weights_dict: Mapping[str, Array],
    path: ParameterPath | None = None,
) -> DACDecoder:
    """Loads an AudioDecoder module from weights.

    This function handles the complete loading of the DAC-style audio decoder, including:
    - first_conv: CausalConv1d with weight normalization
    - decoder_blocks: Multiple AudioDecoderBlock instances
    - final_snake: Snake1d activation
    - final_conv: CausalConv1d with weight normalization

    The PyTorch Decoder stores layers in a single nn.Sequential `model`:
        - model.0: First conv (CausalWNConv1d)
        - model.1 to model.N: DecoderBlocks
        - model.N+1: Final Snake1d
        - model.N+2: Final conv (CausalWNConv1d)
        - model.N+3: Tanh (no weights)

    Expected weight structure (parametrized format):
        model.0.conv.parametrizations.weight.original0/1, model.0.conv.bias
        model.1.block.0.alpha (Snake1d in first DecoderBlock)
        model.1.block.1.conv.parametrizations.weight.original0/1, ... (TransposeConv in first DecoderBlock)
        model.1.block.2.block.0..3 (ResidualUnit in first DecoderBlock)
        ...
        model.N+1.alpha (Final Snake1d)
        model.N+2.conv.parametrizations.weight.original0/1, model.N+2.conv.bias

    Args:
        module: The AudioDecoder module to load weights into.
        weights_dict: Dictionary mapping parameter paths to weight arrays.
        path: Optional base path. If None, uses root path.

    Returns:
        AudioDecoder module with loaded weights.
    """
    if path is None:
        path = ParameterPath()

    # model.0 is the first conv
    first_conv = load_causal_conv1d(module.first_conv, weights_dict, path / "model" / "0" / "conv")

    # model.1 to model.N are decoder blocks
    num_blocks = len(module.decoder_blocks)
    decoder_blocks = tuple(
        load_audio_decoder_block(block, weights_dict, path / "model" / (i + 1))
        for i, block in enumerate(module.decoder_blocks)
    )

    # model.N+1 is final snake (where N = num_blocks)
    final_snake_idx = num_blocks + 1
    final_snake = load_snake1d(module.final_snake, weights_dict, path / "model" / final_snake_idx)

    # model.N+2 is final conv
    final_conv_idx = num_blocks + 2
    final_conv = load_causal_conv1d(module.final_conv, weights_dict, path / "model" / final_conv_idx / "conv")

    return load_parameters(
        lambda m: (m.first_conv, m.decoder_blocks, m.final_snake, m.final_conv),
        module,
        (first_conv, decoder_blocks, final_snake, final_conv),
    )


def load_descript_audio_codec(dac_module: DescriptAudioCodec, state_dict: Mapping[str, Any]) -> DescriptAudioCodec:
    loaded_quantizer = load_downsample_rvq(dac_module.quantizer, state_dict, path=ParameterPath("quantizer"))
    loaded_decoder = load_audio_decoder(dac_module.decoder, state_dict, path=ParameterPath("decoder"))

    return DescriptAudioCodec(
        config=dac_module.config,
        quantizer=loaded_quantizer,
        decoder=loaded_decoder,
    )


def load_fishaudio_text_decoder(
    module: FishAudioTextDecoder,
    weights_dict: Mapping[str, Array],
    decoder_path: ParameterPath | None = None,
) -> FishAudioTextDecoder:
    basepath = ParameterPath() if decoder_path is None else decoder_path
    transformer_slow, readout_slow = load_fish_audio_text_decoding_modules(
        module.transformer_slow,
        module.readout_slow,
        weights_dict,
        fast=False,
    )
    transformer_fast, readout_fast = load_fish_audio_text_decoding_modules(
        module.transformer_fast,
        module.readout_fast,
        weights_dict,
        fast=True,
    )
    embeddings_slow = load_tied_embedding(
        module.embeddings_slow,
        weights_dict,
        basepath / "embeddings",
    )
    embeddings_fast = load_tied_embedding(
        module.embeddings_fast,
        weights_dict,
        basepath / "fast_embeddings",
    )

    codebook_embeddings = load_tied_embedding(
        module.codebook_embeddings,
        weights_dict,
        basepath / "codebook_embeddings",
    )

    if isinstance(module.fast_model_projection, FullPrecisionLinear):
        fast_model_projection = load_linear_and_fuse_scaling(
            module.fast_model_projection,
            weights_dict,
            basepath / "fast_project_in",
        )
        assert isinstance(fast_model_projection, FullPrecisionLinear)
    else:
        fast_model_projection = Identity()

    return load_parameters(
        lambda m: (
            m.embeddings_slow,
            m.transformer_slow,
            m.readout_slow,
            m.embeddings_fast,
            m.transformer_fast,
            m.readout_fast,
            m.codebook_embeddings,
            m.fast_model_projection,
        ),
        module,
        (
            embeddings_slow,
            transformer_slow,
            readout_slow,
            embeddings_fast,
            transformer_fast,
            readout_fast,
            codebook_embeddings,
            fast_model_projection,
        ),
    )


def load_fishaudio_audio_decoder(
    module: DescriptAudioCodec,
    weights_dict: Mapping[str, Array],
    base_path: ParameterPath,
) -> DescriptAudioCodec:
    loaded_quantizer = load_downsample_rvq(module.quantizer, weights_dict, base_path / "quantizer")
    loaded_decoder = load_audio_decoder(module.decoder, weights_dict, base_path / "decoder")

    return load_parameters(lambda m: (m.quantizer, m.decoder), module, (loaded_quantizer, loaded_decoder))


def load_tokenizer_from_fishaudio_tiktoken(
    path_to_tokenizer: Path,
    path_to_special_tokens: Path,
) -> tuple[Tokenizer, FishAudioSpecialInferenceTokens]:
    from tiktoken.core import Encoding as TiktokenEncoding
    from transformers.integrations.tiktoken import convert_tiktoken_to_fast

    def _load_fishaudio_tiktoken_data(
        tiktoken_path: Path,
        special_tokens: dict[str, int],
    ) -> tuple[TiktokenEncoding, FishAudioSpecialInferenceTokens]:
        def load_tiktoken_bpe(tiktoken_bpe_file: Path) -> dict[bytes, int]:
            data = {}
            with open(tiktoken_bpe_file) as token_file:
                for line in token_file.read().splitlines():
                    if not line:
                        continue
                    token, rank = line.split()
                    if token == "=":
                        continue
                    data[base64.b64decode(token)] = int(rank)
            return data

        mergeable_ranks = load_tiktoken_bpe(tiktoken_path)
        special_token_begin = len(mergeable_ranks)
        all_special_tokens_with_ids = {token: special_token_begin + i for i, token in enumerate(special_tokens)}

        semantic_id_to_token_id = {}
        end_idx = 0
        for token in special_tokens:
            if token.startswith("<|semantic:"):
                match_results = re.match(r"<\|semantic:(\d+)\|>", token)
                assert match_results is not None
                idx = int(match_results.group(1))
                semantic_id_to_token_id[idx] = all_special_tokens_with_ids[token]
                end_idx = max(end_idx, idx)

        semantic_begin_id = semantic_id_to_token_id[0]
        semantic_end_id = semantic_id_to_token_id[end_idx]

        tkt_model = TiktokenEncoding(
            name=Path(tiktoken_path).stem,
            pat_str=FISH_TIKTOKEN_PATTERN,
            mergeable_ranks=mergeable_ranks,
            special_tokens=all_special_tokens_with_ids,
        )

        inference_special_tokens = FishAudioSpecialInferenceTokens(
            semantic_begin_id=semantic_begin_id,
            semantic_end_id=semantic_end_id,
            im_end_token_id=all_special_tokens_with_ids[IM_END_TOKEN],
        )

        return tkt_model, inference_special_tokens

    output_temp_dir = tempfile.mkdtemp()
    try:
        if path_to_special_tokens.exists():
            with open(path_to_special_tokens) as f:
                all_special_tokens_with_ids = json.load(f)
        else:
            all_special_tokens_with_ids = {}

        tkt_model, special_inference_tokens = _load_fishaudio_tiktoken_data(
            path_to_tokenizer,
            all_special_tokens_with_ids,
        )

        convert_tiktoken_to_fast(tkt_model, output_temp_dir)
        tokenizer = Tokenizer.from_file(output_temp_dir + "/tokenizer.json")
        return tokenizer, special_inference_tokens
    finally:
        shutil.rmtree(output_temp_dir)



from dataclasses import dataclass, replace
from typing import Any, Self

import jax
from jax import numpy as jnp
from jax import vmap
from jaxtyping import Array, DTypeLike, Float, Int, PRNGKeyArray

from lalamo.common import ParameterTree, require_mapping, require_tree
from lalamo.modules.activations import Identity
from lalamo.modules.audio.fishaudio.fishaudio_common import (
    default_fishaudio_sampling_policy,
)
from lalamo.modules.audio.fishaudio.fishaudio_consts import REPEAT_WINDOW_SIZE, SHORT_LOGITS_SIZE
from lalamo.modules.audio.text_decoder import TTSTextDecoder, TTSTextDecoderConfigBase
from lalamo.modules.common import ForwardPassMode
from lalamo.modules.embedding import TiedEmbedding, TiedEmbeddingConfig
from lalamo.modules.linear import FullPrecisionLinear, FullPrecisionLinearConfig
from lalamo.modules.token_mixers.state.common import State
from lalamo.modules.transformer import Transformer, TransformerConfig
from lalamo.modules.utils import vmap_twice
from lalamo.sampling import SamplingPolicy


@dataclass
class FishAudioTextDecoderResult:
    token_codes: Float[Array, "batch codes"]
    hidden_states: Array | None
    state: State | None


@dataclass(frozen=True)
class FishAudioTextDecoderConfig(TTSTextDecoderConfigBase):
    slow_embeddings_config: TiedEmbeddingConfig
    slow_model_config: TransformerConfig
    slow_readout_config: FullPrecisionLinearConfig

    fast_embeddings_config: TiedEmbeddingConfig
    fast_model_config: TransformerConfig
    fast_readout_config: FullPrecisionLinearConfig

    codebook_embeddings_config: TiedEmbeddingConfig
    fast_model_projection_config: FullPrecisionLinearConfig | None

    semantic_token_begin_id: int
    semantic_token_end_id: int
    im_end_token_id: int
    codebook_size: int
    vocab_size: int
    slow_model_dim: int
    fast_model_dim: int
    num_codebooks: int
    max_seq_len: int

    scale_codebook_embeddings: bool

    precision: DTypeLike

    short_logits_size: int = SHORT_LOGITS_SIZE
    repeat_window_size: int = REPEAT_WINDOW_SIZE

    def empty(self) -> "FishAudioTextDecoder":
        embeddings_slow = self.slow_embeddings_config.empty(self.vocab_size, self.slow_model_dim)
        embeddings_fast = self.fast_embeddings_config.empty(self.codebook_size, self.fast_model_dim)
        codebook_embeddings = self.codebook_embeddings_config.empty(
            self.codebook_size * self.num_codebooks,
            self.slow_model_dim,
        )
        if self.fast_model_projection_config is not None:
            fast_model_projection = self.fast_model_projection_config.empty(
                input_dim=self.slow_model_dim,
                output_dims=(self.fast_model_dim,),
                has_biases=False,
            )
        else:
            fast_model_projection = Identity()
        assert isinstance(embeddings_slow, TiedEmbedding)
        assert isinstance(embeddings_fast, TiedEmbedding)
        assert isinstance(codebook_embeddings, TiedEmbedding)
        return FishAudioTextDecoder(
            self,
            embeddings_slow=embeddings_slow,
            transformer_slow=self.slow_model_config.empty(),
            readout_slow=self.slow_readout_config.empty(
                input_dim=self.slow_model_dim,
                output_dims=(self.vocab_size,),
                has_biases=False,
            ),
            embeddings_fast=embeddings_fast,
            transformer_fast=self.fast_model_config.empty(),
            readout_fast=self.fast_readout_config.empty(
                input_dim=self.fast_model_dim,
                output_dims=(self.codebook_size,),
                has_biases=False,
            ),
            codebook_embeddings=codebook_embeddings,
            fast_model_projection=fast_model_projection,
        )

    def random_init(self, *, key: PRNGKeyArray) -> "FishAudioTextDecoder":
        (
            key_emb_slow,
            key_transformer_slow,
            key_readout_slow,
            key_emb_fast,
            key_transformer_fast,
            key_readout_fast,
            key_emb_codebook,
            key_fast_proj,
        ) = jax.random.split(key, 8)

        embeddings_slow = self.slow_embeddings_config.random_init(
            vocab_size=self.vocab_size,
            model_dim=self.slow_model_dim,
            key=key_emb_slow,
        )
        embeddings_fast = self.fast_embeddings_config.random_init(
            vocab_size=self.codebook_size,
            model_dim=self.fast_model_dim,
            key=key_emb_fast,
        )
        codebook_embeddings = self.codebook_embeddings_config.random_init(
            vocab_size=self.codebook_size * self.num_codebooks,
            model_dim=self.slow_model_dim,
            key=key_emb_codebook,
        )
        if self.fast_model_projection_config is not None:
            fast_model_projection = self.fast_model_projection_config.random_init(
                input_dim=self.slow_model_dim,
                output_dims=(self.fast_model_dim,),
                has_biases=False,
                key=key_fast_proj,
            )
        else:
            fast_model_projection = Identity()

        assert isinstance(embeddings_slow, TiedEmbedding)
        assert isinstance(embeddings_fast, TiedEmbedding)
        assert isinstance(codebook_embeddings, TiedEmbedding)

        return FishAudioTextDecoder(
            self,
            embeddings_slow=embeddings_slow,
            transformer_slow=self.slow_model_config.random_init(key=key_transformer_slow),
            readout_slow=self.slow_readout_config.random_init(
                input_dim=self.slow_model_dim,
                output_dims=(self.vocab_size,),
                has_biases=False,
                key=key_readout_slow,
            ),
            embeddings_fast=embeddings_fast,
            transformer_fast=self.fast_model_config.random_init(key=key_transformer_fast),
            readout_fast=self.fast_readout_config.random_init(
                input_dim=self.fast_model_dim,
                output_dims=(self.codebook_size,),
                has_biases=False,
                key=key_readout_fast,
            ),
            codebook_embeddings=codebook_embeddings,
            fast_model_projection=fast_model_projection,
        )


class FishAudioTextDecoder(TTSTextDecoder[FishAudioTextDecoderConfig]):
    embeddings_slow: TiedEmbedding
    transformer_slow: Transformer
    readout_slow: FullPrecisionLinear

    embeddings_fast: TiedEmbedding
    transformer_fast: Transformer
    readout_fast: FullPrecisionLinear

    codebook_embeddings: TiedEmbedding
    fast_model_projection: FullPrecisionLinear | Identity

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    def export_weights(self) -> ParameterTree:
        if isinstance(self.fast_model_projection, Identity):
            fast_model_proj_weighs = {}
        else:
            fast_model_proj_weighs = self.fast_model_projection.export_weights()

        return {
            "embeddings_slow": self.embeddings_slow.export_weights(),
            "embeddings_fast": self.embeddings_fast.export_weights(),
            "transformer_slow": self.transformer_slow.export_weights(),
            "transformer_fast": self.transformer_fast.export_weights(),
            "readout_slow": self.readout_slow.export_weights(),
            "readout_fast": self.readout_fast.export_weights(),
            "codebook_embeddings": self.codebook_embeddings.export_weights(),
            "fast_model_projection": fast_model_proj_weighs,
        }

    def import_weights(self, weights: ParameterTree) -> Self:
        weights = require_mapping(weights)
        return replace(
            self,
            embeddings_slow=self.embeddings_slow.import_weights(require_tree(weights["embeddings_slow"])),
            embeddings_fast=self.embeddings_fast.import_weights(require_tree(weights["embeddings_fast"])),
            transformer_slow=self.transformer_slow.import_weights(require_tree(weights["transformer_slow"])),
            transformer_fast=self.transformer_fast.import_weights(require_tree(weights["transformer_fast"])),
            readout_slow=self.readout_slow.import_weights(require_tree(weights["readout_slow"])),
            readout_fast=self.readout_fast.import_weights(require_tree(weights["readout_fast"])),
            codebook_embeddings=self.codebook_embeddings.import_weights(require_tree(weights["codebook_embeddings"])),
            fast_model_projection=self.fast_model_projection.import_weights(
                require_tree(weights["fast_model_projection"]),
            )
            if isinstance(self.fast_model_projection, FullPrecisionLinear)
            else Identity(),
        )

    @property
    def semantic_begin_id(self) -> int:
        return self.config.semantic_token_begin_id

    @property
    def semantic_end_id(self) -> int:
        return self.config.semantic_token_end_id

    @property
    def num_codebooks(self) -> int:
        return self.config.num_codebooks

    def embed_slow_model(self) -> Array:
        return jnp.zeros((1, 2, 3))

    def __call__(
        self,
        text_tokens: Int[Array, "batch tokens"],
        sampling_policy: SamplingPolicy,
        key: PRNGKeyArray,
        input_pos: Int[Array, "batch tokens"] | None = None,
        state: State | None = None,
    ) -> FishAudioTextDecoderResult:
        batch_size, seq_length = text_tokens.shape
        if input_pos is None:
            input_pos = jnp.arange(seq_length)[None, :]

        text_and_codebooks = jnp.zeros(
            (batch_size, self.config.num_codebooks + 1, seq_length),
            dtype=text_tokens.dtype,
        )
        # NOTE: the rest of codebook lines should be filled in case audio prompt is used, but
        # ignore it for now
        text_and_codebooks = text_and_codebooks.at[:, 0, :].set(text_tokens)

        embeddings = self.embed(text_and_codebooks)

        codes, updated_state = decode_next_token(
            model=self,
            x=embeddings,
            state_slow=state,
            input_pos=input_pos,
            sampling_policy=sampling_policy,
            previous_tokens=None,
            key=key,
        )

        return FishAudioTextDecoderResult(token_codes=codes, hidden_states=None, state=updated_state)

    def embed(
        self,
        inp: Int[Array, "batch codebooks tokens"],
        apply_codebook_embeddings: bool = False,
    ) -> Float[Array, "batch tokens embedding"]:
        """
        apply_codebook_embeddings argument should be set to 'True' if audio-prompt is used. In this
        case we expect codebook lines [1:-1] to be filled with something meaningful
        """

        vq_masks = (inp[:, 0] >= self.semantic_begin_id) & (inp[:, 0] <= self.semantic_end_id)
        embeddings = self.embeddings_slow.embed(inp[:, 0])

        if apply_codebook_embeddings or jnp.any(vq_masks):
            _, _, seq_length = inp.shape
            codebook_offsets = (jnp.arange(self.config.num_codebooks) * self.config.codebook_size).reshape(-1, 1)
            codebook_offsets = jnp.tile(codebook_offsets, (1, seq_length))
            codebook_embeds = vmap(self.codebook_embeddings.embed)(inp[:, 1:, :] + codebook_offsets)

            vq_embeds_sum = codebook_embeds.sum(axis=1)
            vq_embeds_sum = vq_embeds_sum.at[~vq_masks].set(0)
            embeddings = embeddings + vq_embeds_sum

        if self.config.scale_codebook_embeddings:
            # Expand vq_masks to match x's shape
            vq_masks_expanded = jnp.expand_dims(vq_masks, axis=-1)
            vq_masks_expanded = jnp.broadcast_to(vq_masks_expanded, embeddings.shape)
            embeddings = jnp.where(vq_masks_expanded, embeddings / jnp.sqrt(self.config.num_codebooks + 1), embeddings)
            assert isinstance(embeddings, Array)

        return embeddings

    def decode_utterance(
        self,
        text_tokens: Int[Array, "batch tokens"],
        sampling_policy: SamplingPolicy | None = None,
        key: PRNGKeyArray | None = None,
    ) -> Int[Array, "num_codebooks tokens"]:
        """
        Generate semantic tokens for a full utterance given text tokens in an autoregressive
        generation loop. Processing text tokens through the slow transformer and generating
        codebook tokens until the end token is reached or max sequence length is exceeded.
        Returns:
            Generated codebook tokens for DAC codec
        """

        batch_size, prompt_length = text_tokens.shape
        assert batch_size == 1, "Only batch_size=1 is supported"

        codebook_dim = 1 + self.config.num_codebooks
        max_seq_len = self.config.max_seq_len

        if prompt_length >= max_seq_len:
            raise ValueError(f"Input sequence length {prompt_length} exceeds max_seq_len {max_seq_len}")

        if sampling_policy is None:
            sampling_policy = default_fishaudio_sampling_policy()
        if key is None:
            key = jax.random.PRNGKey(123)

        max_new_tokens = max_seq_len - prompt_length

        # Prepare prompt: text tokens in first row
        # Rest of codebook rows are zeros until we start using audio-embeddings for explicit style
        prompt = jnp.zeros((batch_size, codebook_dim, prompt_length), dtype=text_tokens.dtype)
        prompt = prompt.at[:, 0, :].set(text_tokens)

        # Initialize sequence buffer to store generated tokens
        seq = jnp.zeros((codebook_dim, max_seq_len), dtype=jnp.int32)
        seq = seq.at[:, :prompt_length].set(prompt[0])

        # Track previous tokens for repetition penalty (windowed)
        previous_tokens = jnp.zeros((codebook_dim, max_seq_len), dtype=jnp.int32)

        input_pos = jnp.arange(prompt_length)[None, :]
        embeddings = self.embed(prompt)
        first_codes, state_slow = decode_next_token(
            model=self,
            x=embeddings,
            state_slow=None,
            input_pos=input_pos,
            sampling_policy=sampling_policy,
            key=key,
            previous_tokens=None,
        )

        seq = seq.at[:, prompt_length].set(first_codes[0])
        previous_tokens = previous_tokens.at[:, 0].set(first_codes[0])

        if first_codes[0, 0] == self.config.im_end_token_id:
            codes = seq[1:, prompt_length : prompt_length + 1]
            return codes

        cur_token = first_codes
        generated_count = 1
        for i in range(1, max_new_tokens):
            cur_token_expanded = cur_token.reshape(batch_size, codebook_dim, 1)

            # Get windowed previous tokens for repetition penalty
            win_size = self.config.repeat_window_size
            if i < win_size:
                window = previous_tokens[:, :win_size]
            else:
                window = previous_tokens[:, i - win_size : i]

            embeddings = self.embed(cur_token_expanded)

            input_pos = jnp.array([[prompt_length + i - 1]])

            if key is not None:
                key, subkey = jax.random.split(key)
            else:
                subkey = None

            next_codes, state_slow = decode_next_token(
                model=self,
                x=embeddings,
                state_slow=state_slow,
                input_pos=input_pos,
                sampling_policy=sampling_policy,
                key=subkey,
                previous_tokens=window,
            )

            seq = seq.at[:, prompt_length + i].set(next_codes[0])
            previous_tokens = previous_tokens.at[:, i].set(next_codes[0])
            generated_count += 1

            if next_codes[0, 0] == self.config.im_end_token_id:
                break

            cur_token = next_codes

        # Extract codebook codes (exclude text token row and prompt, exclude last token which is end token)
        codes = seq[1:, prompt_length : prompt_length + generated_count - 1]
        assert jnp.all(codes >= 0), "Negative code found"

        return codes


def decode_next_token(
    model: FishAudioTextDecoder,
    x: Array,
    state_slow: State | None,
    input_pos: Array,
    sampling_policy: SamplingPolicy,
    key: PRNGKeyArray,
    previous_tokens: Array | None = None,  # noqa: ARG001, reserved for future when repetition penalty is done
) -> tuple[Int[Array, "batch codes"], State | None]:
    batch_size = x.shape[0]
    assert batch_size == 1, "Batching not supported yet"

    slow_model_result = model.transformer_slow(
        inner_features=x,
        token_positions=input_pos,
        state=state_slow,
        return_updated_state=True,
        return_layer_results=True,
        return_positional_embeddings=False,
        lengths_without_padding=None,
        forward_pass_mode=ForwardPassMode.MULTI_TOKEN,
        forward_pass_config=None,
    )
    assert slow_model_result.layer_results is not None
    hidden_states = slow_model_result.layer_results[-1].outputs[:, -1:]
    (hidden_states,) = vmap(model.fast_model_projection)(hidden_states)
    hidden_states = hidden_states.reshape(hidden_states.shape[0], 1, -1)

    (logits,) = vmap_twice(model.readout_slow)(slow_model_result.outputs)

    n_codes = model.num_codebooks + 1

    codebooks = jnp.zeros((batch_size, n_codes), dtype=jnp.int32)
    codebooks = codebooks.at[0, 0].set(vmap(lambda x: sampling_policy(x, key=key))(logits[:, -1, :])[0])
    first_fast_code = jnp.array([codebooks[0, 0] - model.semantic_begin_id])
    first_fast_code = first_fast_code.at[first_fast_code < 0].set(0)
    codebooks = codebooks.at[0, 1].set(first_fast_code[0])

    state_fast = model.transformer_fast.init_static_state(batch_size, n_codes)

    input_pos_fast = jnp.zeros((batch_size, 1), dtype=jnp.int32)
    fast_first_result = model.transformer_fast(
        inner_features=hidden_states,
        token_positions=input_pos_fast,
        state=state_fast,
        return_updated_state=True,
        return_layer_results=False,
        return_positional_embeddings=False,
        lengths_without_padding=None,
        forward_pass_mode=ForwardPassMode.SINGLE_TOKEN,
        forward_pass_config=None,
    )
    state_fast = fast_first_result.updated_state

    embedded_logits = model.embeddings_fast.embed(first_fast_code)

    def loop_iteration(
        iteration_state: tuple[State | None, Array, Any],
        index: jnp.int32,
    ) -> tuple[tuple[State | None, Array, Any], None]:
        transformer_state, logits, codebooks = iteration_state
        logits = logits.reshape(logits.shape[0], 1, -1)
        input_pos_fast = jnp.array([index - 1])[None, :]
        fast_result = model.transformer_fast(
            inner_features=logits,
            token_positions=input_pos_fast,
            state=transformer_state,
            return_updated_state=True,
            return_layer_results=False,
            return_positional_embeddings=False,
            lengths_without_padding=None,
            forward_pass_mode=ForwardPassMode.SINGLE_TOKEN,
            forward_pass_config=None,
        )
        (fast_logits,) = vmap_twice(model.readout_fast)(fast_result.outputs)
        new_state = fast_result.updated_state

        short_logits = fast_logits[:, :, : model.config.short_logits_size]
        code = vmap(lambda x: sampling_policy(x, key=key))(short_logits[:, -1, :])

        new_logits = model.embeddings_fast.embed(code)
        codebooks = codebooks.at[0, index].set(code[0])
        return (new_state, new_logits, codebooks), None

    scan_result, _ = jax.lax.scan(
        loop_iteration,
        (state_fast, embedded_logits, codebooks),
        jnp.arange(2, n_codes, dtype=jnp.int32),
    )
    _, _, codebooks = scan_result

    return codebooks, slow_model_result.updated_state



from dataclasses import dataclass
from enum import StrEnum


class AudioEncoding(StrEnum):
    PCM = "pcm"
    ULAW = "uLAW"
    ALAW = "aLAW"


@dataclass(frozen=True)
class AudioRenderingSettings:
    samplerate: int
    output_channels: int
    bitwidth: int
    encoding: AudioEncoding



import struct
from array import array
from collections.abc import Callable, Iterable
from dataclasses import dataclass, field
from itertools import chain, repeat, tee
from math import exp
from typing import Self

import xxhash

from .common import Speculator


def padded_sliding_window(seq: Iterable[int], size: int, pad: int) -> Iterable[tuple[int, ...]]:
    seqs = tee(seq, size)
    pads = tuple(repeat(pad, size - i) for i in range(size))
    padded_seqs = tuple(chain(pad, seq) for pad, seq in zip(pads, seqs, strict=True))
    return zip(*padded_seqs, strict=False)


def softmax(logits: Iterable[float]) -> list[float]:
    logits = list(logits)
    log_max = max(logits)
    exp_logs = [exp(logit - log_max) for logit in logits]
    exp_log_sum = sum(exp_logs)
    return [exp_log / exp_log_sum for exp_log in exp_logs]


def update_probs(probs: dict[int, float], sample: dict[int, float], new_count: int, top_k: int) -> None:
    """Online mean update with top-K truncation. Mutates probs in-place."""
    decay = (new_count - 1) / new_count
    inv_n = 1.0 / new_count
    for k in probs:
        probs[k] *= decay
    for k, v in sample.items():
        probs[k] = probs.get(k, 0.0) + v * inv_n
    if len(probs) > top_k:
        keep = dict(sorted(probs.items(), key=lambda x: (-x[1], x[0]))[:top_k])
        probs.clear()
        probs.update(keep)
    total = sum(probs.values())
    if total > 0:
        inv_total = 1.0 / total
        for k in probs:
            probs[k] *= inv_total


def seqhash(tokens: Iterable[int], size: int) -> tuple[int, int]:
    tokens = list(tokens)
    packed = struct.pack("<" + "I" * len(tokens), *tokens) if tokens else b""
    h = xxhash.xxh3_64_intdigest(packed)
    return h % size, h // size


def _discount_and_interpolate(
    dist: dict[int, float],
    lower_dist: dict[int, float],
    discount: float,
) -> dict[int, float]:
    num_nonzero = sum(1 for v in dist.values() if v > 0)
    freed_mass = min(discount * num_nonzero, 0.5)
    smoothed = {k: max(v - discount, 0.0) for k, v in dist.items()}
    for token, prob in lower_dist.items():
        smoothed[token] = smoothed.get(token, 0.0) + freed_mass * prob
    return smoothed


def _write_top_k(dist: dict[int, float], keys: array, values: array, idx: int, ngram_k: int) -> None:
    top = sorted(dist.items(), key=lambda x: (-x[1], x[0]))[:ngram_k]
    total = sum(v for _, v in top)
    if total <= 0:
        return
    new_keys = array("I", repeat(0, ngram_k))
    new_values = array("f", repeat(0.0, ngram_k))
    for i, (token_id, prob) in enumerate(top):
        new_keys[i] = token_id
        new_values[i] = prob / total
    k_start = idx * ngram_k
    memoryview(keys)[k_start : k_start + ngram_k] = new_keys
    memoryview(values)[k_start : k_start + ngram_k] = new_values


class ExactBucket:
    __slots__ = ("count", "probs")

    def __init__(self) -> None:
        self.count: int = 0
        self.probs: dict[int, float] = {}


@dataclass(frozen=True, eq=False)
class TaggedNGramTable:
    hashtable_size: int
    ngram_k: int
    ngram_n: int
    ngram_pad: int

    # Training phase: exact data (interior mutability)
    _exact_data: dict[tuple[int, int], ExactBucket] = field(default_factory=dict, repr=False)
    _context_tokens: dict[tuple[int, int], tuple[int, ...]] = field(default_factory=dict, repr=False)

    # Inference phase: tagged hash table (populated by compress())
    _tags: array = field(default_factory=lambda: array("Q"), repr=False)
    _keys: array = field(default_factory=lambda: array("I"), repr=False)
    _values: array = field(default_factory=lambda: array("f"), repr=False)
    _counts: array = field(default_factory=lambda: array("I"), repr=False)
    _compressed: list[bool] = field(default_factory=lambda: [False], repr=False)

    @classmethod
    def init(cls, hashtable_size: int, ngram_k: int, ngram_n: int, ngram_pad: int = 2**32 - 1) -> Self:
        return cls(
            hashtable_size=hashtable_size,
            ngram_k=ngram_k,
            ngram_n=ngram_n,
            ngram_pad=ngram_pad,
        )

    def _context_for(self, seq: Iterable[int]) -> list[int]:
        seq = list(seq)
        ngram_ctx = self.ngram_n - 1
        if ngram_ctx > 0:
            return [*repeat(self.ngram_pad, max(ngram_ctx - len(seq), 0)), *seq[-ngram_ctx:]]
        return []

    def train(self, token_ids: Iterable[int], token_logits: Iterable[dict[int, float]]) -> None:
        ngram_ctx = self.ngram_n - 1
        if ngram_ctx > 0:
            contexts = padded_sliding_window(token_ids, ngram_ctx, self.ngram_pad)
        else:
            contexts = repeat(())

        for ctx, cur_logits in zip(contexts, token_logits, strict=False):
            ctx_tuple = tuple(ctx)
            idx, tag = seqhash(ctx_tuple, self.hashtable_size)
            key = (idx, tag)
            if key not in self._context_tokens:
                self._context_tokens[key] = ctx_tuple
            bucket = self._exact_data.get(key)

            if bucket is None:
                bucket = ExactBucket()
                self._exact_data[key] = bucket

            bucket.count += 1
            sample = dict(zip(cur_logits.keys(), softmax(cur_logits.values()), strict=True))
            update_probs(bucket.probs, sample, bucket.count, self.ngram_k)

    def compress(self) -> None:
        tags = array("Q", repeat(0, self.hashtable_size))
        keys = array("I", repeat(0, self.hashtable_size * self.ngram_k))
        values = array("f", repeat(0.0, self.hashtable_size * self.ngram_k))
        bucket_counts: array = array("I", repeat(0, self.hashtable_size))

        for (idx, tag), bucket in self._exact_data.items():
            if bucket.count > bucket_counts[idx]:
                bucket_counts[idx] = bucket.count
                tags[idx] = tag
                _write_top_k(bucket.probs, keys, values, idx, self.ngram_k)

        self._tags[:] = tags
        self._keys[:] = keys
        self._values[:] = values
        self._counts[:] = bucket_counts
        self._compressed[0] = True

    def probs(self, seq: Iterable[int]) -> dict[int, float]:
        assert self._compressed[0], "call compress() before probs()"
        ctx = self._context_for(seq)
        idx, tag = seqhash(ctx, self.hashtable_size)

        if self._tags[idx] != tag:
            return {}

        k_start = idx * self.ngram_k
        k_end = k_start + self.ngram_k
        return {k: v for k, v in zip(self._keys[k_start:k_end], self._values[k_start:k_end], strict=True) if v > 0.0}

    def exact_dist_for(self, seq: Iterable[int]) -> dict[int, float] | None:
        """Look up exact (pre-compress) distribution for a context."""
        key = seqhash(self._context_for(seq), self.hashtable_size)
        bucket = self._exact_data.get(key)
        return dict(bucket.probs) if bucket is not None else None

    def apply_smoothing(
        self,
        discount: float,
        lower_order_fn: Callable[[tuple[int, ...]], dict[int, float]],
    ) -> None:
        """Re-smooth compressed buckets using lower-order interpolation."""
        for (idx, tag), bucket in self._exact_data.items():
            if self._tags[idx] != tag:
                continue
            ctx_tokens = self._context_tokens.get((idx, tag))
            if ctx_tokens is None:
                continue
            lower_dist = lower_order_fn(ctx_tokens)
            smoothed = _discount_and_interpolate(bucket.probs, lower_dist, discount)
            _write_top_k(smoothed, self._keys, self._values, idx, self.ngram_k)

    def serialize(self) -> bytes:
        assert self._compressed[0], "call compress() before serialize()"
        hdr = struct.pack("<4I", self.hashtable_size, self.ngram_k, self.ngram_n, self.ngram_pad)
        return hdr + bytes(self._tags) + bytes(self._keys) + bytes(self._values) + bytes(self._counts)

    @classmethod
    def deserialize(cls, blob: bytes) -> Self:
        offset = 16
        hashtable_size, ngram_k, ngram_n, ngram_pad = struct.unpack("<4I", blob[:offset])

        tag_len = 8 * hashtable_size
        tags = array("Q", blob[offset : offset + tag_len])
        offset += tag_len

        kv_len = 4 * ngram_k * hashtable_size
        keys = array("I", blob[offset : offset + kv_len])
        offset += kv_len

        values = array("f", blob[offset : offset + kv_len])
        offset += kv_len

        count_len = 4 * hashtable_size
        counts = array("I", blob[offset : offset + count_len])
        offset += count_len

        instance = cls(
            hashtable_size=hashtable_size,
            ngram_k=ngram_k,
            ngram_n=ngram_n,
            ngram_pad=ngram_pad,
        )
        instance._tags[:] = tags
        instance._keys[:] = keys
        instance._values[:] = values
        instance._counts[:] = counts
        instance._compressed[0] = True
        return instance


def _deserialize_tagged_tables(blob: bytes, offset: int, max_order: int) -> tuple[list[TaggedNGramTable], int]:
    tables: list[TaggedNGramTable] = []
    for _ in range(max_order):
        (length,) = struct.unpack("<Q", blob[offset : offset + 8])
        offset += 8
        tables.append(TaggedNGramTable.deserialize(blob[offset : offset + length]))
        offset += length
    return tables, offset


@dataclass(frozen=True, eq=False)
class NGramSpeculator(Speculator):
    max_order: int
    tables: tuple[TaggedNGramTable, ...]
    discount: float = 0.002

    @classmethod
    def init(cls, hashtable_size: int, ngram_k: int, max_order: int = 4, discount: float = 0.002) -> Self:
        tables = tuple(TaggedNGramTable.init(hashtable_size, ngram_k, n) for n in range(1, max_order + 1))
        return cls(max_order, tables, discount)

    def train(self, token_ids: Iterable[int], token_logits: Iterable[dict[int, float]]) -> None:
        ids = list(token_ids)
        logits = list(token_logits)
        for table in self.tables:
            table.train(ids, logits)

    def _get_best_lower_order_dist(self, ctx_tokens: tuple[int, ...], max_lower_order: int) -> dict[int, float]:
        """Find the highest available lower-order distribution for backoff."""
        for order in range(max_lower_order, 0, -1):
            table = self.tables[order - 1]
            shorter_ctx = ctx_tokens[-(order - 1) :] if order > 1 else ()
            dist = table.exact_dist_for(shorter_ctx)
            if dist is not None:
                return dist
        return {}

    def compress(self) -> None:
        for table in self.tables:
            table.compress()

        for order_idx in range(1, len(self.tables)):
            table = self.tables[order_idx]
            table.apply_smoothing(
                self.discount,
                lambda ctx, oi=order_idx: self._get_best_lower_order_dist(ctx, oi),
            )

    def probs(self, seq: Iterable[int]) -> dict[int, float]:
        seq = list(seq)
        for table in reversed(self.tables):
            result = table.probs(seq)
            if result:
                return result
        return {}

    def serialize(self) -> bytes:
        parts = [struct.pack("<If", self.max_order, self.discount)]
        for table in self.tables:
            table_bytes = table.serialize()
            parts.append(struct.pack("<Q", len(table_bytes)))
            parts.append(table_bytes)
        return b"".join(parts)

    @classmethod
    def deserialize(cls, blob: bytes) -> Self:
        offset = 0
        max_order, discount = struct.unpack("<If", blob[offset : offset + 8])
        offset += 8
        tables, _ = _deserialize_tagged_tables(blob, offset, max_order)
        return cls(max_order, tuple(tables), discount)



import json
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

import equinox as eqx
import jax
import numpy as np
from jax import Array
from jax import numpy as jnp
from jaxtyping import DTypeLike, Float, Int, PRNGKeyArray
from tokenizers import Tokenizer

from lalamo.audio.audio_rendering import AudioEncoding, AudioRenderingSettings
from lalamo.audio.tts_message_processor import (
    TTSMessage,
    TTSMessageProcessor,
    TTSMessageProcessorConfig,
)
from lalamo.modules import TTSModel, config_converter
from lalamo.modules.audio.fishaudio.fishaudio_common import (
    default_fishaudio_sampling_policy,
)
from lalamo.modules.audio.fishaudio.fishaudio_consts import (
    DEFAULT_FISH_AUDIO_REPETITION_PENALTY,
    DEFAULT_FISHAUDIO_RANDOM_SEED,
)
from lalamo.modules.audio.fishaudio.fishaudio_text_decoding import (
    FishAudioTextDecoder,
)
from lalamo.modules.audio.text_to_speech import (
    DEFAULT_TTS_REPETITION_PENALTY,
    DEFAULT_TTS_SAMPLING_POLICY,
    TTSConfig,
)
from lalamo.safetensors import safe_read
from lalamo.sampling import SamplingPolicy

from .common import ParameterTree, unflatten_parameters

__all__ = ["TTSGenerationResult", "TTSGenerator", "TTSGeneratorConfig"]


@dataclass(frozen=True)
class TTSGeneratorConfig:
    tts_config: TTSConfig
    message_processor_config: TTSMessageProcessorConfig


@dataclass
class TTSGenerationResult:
    audio: np.ndarray
    audio_params: AudioRenderingSettings


class TTSGenerator(eqx.Module):
    config: TTSGeneratorConfig
    tts_model: TTSModel

    message_processor: TTSMessageProcessor = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.tts_model.text_decoder.activation_precision

    def export_weights(self) -> ParameterTree[Array]:
        return {
            "text_decoder": self.tts_model.text_decoder.export_weights(),
            "audio_decoder": self.tts_model.audio_decoder.export_weights(),
            "vocoder": self.tts_model.vocoder.export_weights(),
        }

    def tokenize_text(self, messages: Iterable[TTSMessage]) -> Int[Array, " batch tokens"]:
        text_tokens = self.message_processor.tokenize_request(messages)
        return jnp.asarray(text_tokens)[None, :]

    def decode_text(
        self,
        text_tokens: Array,
        sampling_policy: SamplingPolicy,
        repetition_penalty: float,  # noqa: ARG002, reserved for near future
        random_key: PRNGKeyArray | None = None,
    ) -> Array:
        random_key = jax.random.PRNGKey(123) if random_key is None else random_key
        return self.tts_model.text_decoder.decode_utterance(
            text_tokens,
            sampling_policy=sampling_policy,
            key=random_key,
        )

    def decode_audio(self, semantic_tokens: Array) -> Array:
        return self.tts_model.audio_decoder.audio_from_codes(semantic_tokens)

    def generate_waveform(self, audio_features: Array) -> Array:
        return self.tts_model.vocoder(audio_features)

    def get_generated_audio_params(self) -> AudioRenderingSettings:
        # NOTE: think if this could be moved to config level or made mandatory via abstract
        return AudioRenderingSettings(
            samplerate=self.tts_model.audio_decoder.samplerate,
            output_channels=1,
            bitwidth=16,
            encoding=AudioEncoding.PCM,
        )

    def generate_speech(
        self,
        messages: Iterable[TTSMessage],
        sampling_policy: SamplingPolicy = DEFAULT_TTS_SAMPLING_POLICY,
        repetition_penalty: float = DEFAULT_TTS_REPETITION_PENALTY,
        random_key: PRNGKeyArray | None = None,
    ) -> TTSGenerationResult:
        text_tokens = self.tokenize_text(messages)

        semantic_tokens = self.decode_text(
            text_tokens,
            sampling_policy=sampling_policy,
            repetition_penalty=repetition_penalty,
            random_key=random_key,
        )

        audio_features = self.decode_audio(semantic_tokens)

        audio_waveform = self.tts_model.vocoder(audio_features)

        audio_settings = self.get_generated_audio_params()

        return TTSGenerationResult(audio=np.array(audio_waveform), audio_params=audio_settings)

    @classmethod
    def load_model(cls, path: Path | str) -> "TTSGenerator":
        if isinstance(path, str):
            path = Path(path)
        with open(path / "config.json") as config_file:
            config_json = json.load(config_file)

        config = config_converter.structure(config_json["model_config"], TTSGeneratorConfig)
        assert isinstance(config, TTSGeneratorConfig)
        with Path(path / "model.safetensors").open("rb") as fd:
            _, weights_dict = safe_read(fd)
            weights = unflatten_parameters(weights_dict)
            model = config.tts_config.empty().import_weights(weights)
        tokenizer = Tokenizer.from_file(str(path / "tokenizer.json"))
        message_processor = TTSMessageProcessor(config.message_processor_config, tokenizer)
        return TTSGenerator(
            config=config,
            tts_model=model,
            message_processor=message_processor,
        )


class FishAudioTTSGenerator(TTSGenerator):
    def decode_text(
        self,
        text_tokens: Int[Array, "batch sequence"],
        sampling_policy: SamplingPolicy | None = None,
        repetition_penalty: float = DEFAULT_FISH_AUDIO_REPETITION_PENALTY,  # noqa: ARG002, reserved for near future
        random_key: PRNGKeyArray | None = None,
    ) -> Int[Array, "num_codebooks sequence"]:
        assert isinstance(self.tts_model.text_decoder, FishAudioTextDecoder)

        sampling_policy = sampling_policy if sampling_policy is not None else default_fishaudio_sampling_policy()

        random_key = jax.random.PRNGKey(DEFAULT_FISHAUDIO_RANDOM_SEED) if random_key is None else random_key
        return self.tts_model.text_decoder.decode_utterance(
            text_tokens,
            sampling_policy=sampling_policy,
            key=random_key,
        )

    def decode_audio(
        self,
        semantic_tokens: Int[Array, "batch n_codebooks sequence"],
    ) -> Float[Array, "batch audio_samples 1"]:
        return super().decode_audio(semantic_tokens)

    def generate_waveform(self, audio_features: Array) -> Array:
        return super().generate_waveform(audio_features)

    def get_generated_audio_params(self) -> AudioRenderingSettings:
        return AudioRenderingSettings(
            samplerate=self.tts_model.audio_decoder.samplerate,
            output_channels=1,
            bitwidth=16,
            encoding=AudioEncoding.PCM,
        )



import string

import numpy as np
from tokenizers import Tokenizer, pre_tokenizers
from tokenizers.models import WordLevel

__all__ = ["dummy_char_level_tokenizer_config", "play_mono_audio"]

DEFAULT_SAMPLERATE: int = 44100


def play_mono_audio(audio: np.ndarray, samplerate: int, audio_chunk_size: int = 1024) -> None:
    import pyaudio  # type: ignore[reportMissingImports], pyaudio is optional to install,

    if audio.dtype not in [np.float32, np.float16, np.float64]:
        raise ValueError("Input audio datatype is expected to be a floating point")
    (n_samples,) = audio.shape
    # better to clip then to overflow
    audio = np.clip(audio, -1.0, 1.0)
    # very dumb conversion to PCM16
    pcm_audio = (audio * np.iinfo(np.int16).max).astype(np.int16)

    audio_chunk_size = 1024
    num_chunks = int(np.ceil(n_samples / audio_chunk_size))

    # actual size of each chunk might not be exactly 'audio_chunk_size' but not critical here
    chunks = np.array_split(pcm_audio, num_chunks)

    audio_interface = pyaudio.PyAudio()
    output_stream = audio_interface.open(
        format=audio_interface.get_format_from_width(pcm_audio.dtype.itemsize),
        channels=1,
        rate=samplerate,
        output=True,
    )

    for chunk in chunks:
        output_stream.write(chunk.tobytes())


def dummy_char_level_tokenizer_config() -> str:
    chars = list(string.ascii_lowercase + string.ascii_uppercase + string.digits + string.punctuation + " ")
    vocab = {char: idx for idx, char in enumerate(chars)}
    vocab["[UNK]"] = len(vocab)
    vocab["[PAD]"] = len(vocab)

    tokenizer = Tokenizer(WordLevel(vocab=vocab, unk_token="[UNK]"))
    tokenizer.pre_tokenizer = pre_tokenizers.Split("", behavior="isolated")  # type: ignore[assignment]

    return tokenizer.to_str()



"""Default constants for NanoCodec modules."""

from jax import numpy as jnp

DEFAULT_LEAKY_RELU_NEGATIVE_SLOPE: float = 0.01
DEFAULT_AUDIO_DECODER_INPUT_CONV_SIZE: int = 7
DEFAULT_AUDIO_DECODER_OUTPUT_CONV_SIZE: int = 3
DEFAULT_AUDIO_DECODER_RESBLOCK_KERNEL_SIZES = (3, 7, 11)
DEFAULT_AUDIO_DECODER_RESBLOCK_DILATIONS = (1, 3, 5)
DEFAULT_FSQ_EPS: float = 1e-3

DEFAULT_NANOCODEC_PRECISION = jnp.float32



import json
from abc import ABC, abstractmethod
from collections.abc import Iterable
from dataclasses import dataclass, replace
from pathlib import Path
from typing import Self

import equinox as eqx
import jax
from jax import Array
from jax import numpy as jnp
from jaxtyping import PRNGKeyArray
from tokenizers import Tokenizer

from lalamo.common import DTypeLike, ParameterTree, unflatten_parameters
from lalamo.message_processor import Message, MessageProcessor, MessageProcessorConfig, UserMessage
from lalamo.modules import Classifier, Decoder, LalamoModule, config_converter
from lalamo.modules.classifier import ClassifierConfig, ClassifierResult
from lalamo.modules.decoder import DecoderConfig, DecoderResult
from lalamo.safetensors import safe_read

__all__ = [
    "BatchSizeInfo",
    "BatchSizesComputedEvent",
    "TextModel",
    "TextModelConfig",
    "split_into_rolling_keys",
]


def split_into_rolling_keys(key: PRNGKeyArray, num: int = 2) -> PRNGKeyArray:
    """Equivalent to jax.random.split(), but uses lax.scan to produce keys
    sequentially (rolling), so that keys share consistent prefixes."""
    if not isinstance(num, int):
        raise TypeError(f"Expected integer 'num', got {type(num).__name__}")
    if num < 0:
        raise ValueError(f"Expected non-negative 'num', got {num}")

    def split_once(carry: PRNGKeyArray, _: None) -> tuple[PRNGKeyArray, PRNGKeyArray]:
        next_carry, sample_key = jax.random.split(carry)
        return next_carry, sample_key

    _, keys = jax.lax.scan(split_once, key, xs=None, length=num)
    return keys


@dataclass(frozen=True)
class InferenceConfig:
    max_output_length: int = 8192
    padded_length: int = 8192
    num_top_logits_to_return: int | None = None
    batch_size: int | None = None


@dataclass(frozen=True)
class BatchSizeInfo:
    prefix_length: int
    num_elements: int
    batch_size: int


@dataclass(frozen=True)
class BatchSizesComputedEvent:
    batch_sizes: tuple[BatchSizeInfo, ...]


@dataclass(frozen=True)
class TextModelConfig[ConfigT: ClassifierConfig | DecoderConfig](ABC):
    model_config: ConfigT
    message_processor_config: MessageProcessorConfig

    @abstractmethod
    def init(
        self,
        model: LalamoModule,
        message_processor: MessageProcessor,
    ) -> LalamoModule[Self]: ...

    @classmethod
    def load_model(cls, path: Path | str) -> LalamoModule[Self]:
        if isinstance(path, str):
            path = Path(path)
        with open(path / "config.json") as config_file:
            config_json = json.load(config_file)
        config = config_converter.structure(config_json["model_config"], cls)
        with Path(path / "model.safetensors").open("rb") as fd:
            _, weights_dict = safe_read(fd)
            weights = unflatten_parameters(weights_dict)
            model = config.model_config.empty().import_weights(weights)  # type: ignore
        tokenizer = Tokenizer.from_file(str(path / "tokenizer.json"))
        message_processor = MessageProcessor(config.message_processor_config, tokenizer)
        return config.init(model, message_processor)


class TextModel[ConfigT, ModelT: Decoder | Classifier](LalamoModule[ConfigT]):
    model: ModelT
    message_processor: MessageProcessor = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.model.activation_precision

    def export_weights(self) -> ParameterTree:
        return self.model.export_weights()  # type: ignore

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        return replace(
            self,
            model=self.model.import_weights(weights),  # type: ignore
        )

    def record_trace(self, messages: Iterable[Message] | None = None) -> ClassifierResult | DecoderResult:
        if messages is None:
            messages = [UserMessage("Tell me about London")]

        token_ids = jnp.array(self.message_processor.tokenize_request(messages))[None, :]
        _, num_tokens = token_ids.shape
        token_positions = jnp.arange(num_tokens)[None, :]
        return self.model(token_ids=token_ids, token_positions=token_positions, return_activation_trace=True)



from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Self

import jax
from jaxtyping import Array, DTypeLike, PRNGKeyArray

from lalamo.common import ParameterTree, require_tree
from lalamo.modules.common import LalamoModule, register_config_union
from lalamo.sampling import SamplingPolicy, make_policy

from .audio_decoder import TTSAudioDecoder
from .fishaudio.fishaudio_audio_decoding import DescriptAudioCodecConfig
from .fishaudio.fishaudio_text_decoding import FishAudioTextDecoderConfig
from .nanocodec.audio_decoding import NanoCodecConfig
from .nanocodec.stub_text_decoder import StubTextDecoderConfig
from .text_decoder import TTSTextDecoder
from .vocoders import Vocoder, VocoderConfig

DEFAULT_TTS_SAMPLING_POLICY: SamplingPolicy = make_policy(temperature=0.3, top_p=0.9)
DEFAULT_TTS_REPETITION_PENALTY: float = 1.1


TTSAudioDecoderConfig = DescriptAudioCodecConfig | NanoCodecConfig
register_config_union(TTSAudioDecoderConfig)

TTSTextDecoderConfig = FishAudioTextDecoderConfig | StubTextDecoderConfig
register_config_union(TTSTextDecoderConfig)


@dataclass(frozen=True)
class TTSConfig:
    text_decoder_config: TTSTextDecoderConfig
    audio_decoder_config: TTSAudioDecoderConfig
    vocoder_config: VocoderConfig

    activation_precision: DTypeLike

    def empty(self) -> "TTSModel":
        text_decoder = self.text_decoder_config.empty()
        audio_decoder = self.audio_decoder_config.empty()
        vocoder = self.vocoder_config.empty()
        return TTSModel(config=self, text_decoder=text_decoder, audio_decoder=audio_decoder, vocoder=vocoder)

    def random_init(self, key: PRNGKeyArray) -> "TTSModel":
        key_text, key_audio = jax.random.split(key)
        text_decoder = self.text_decoder_config.random_init(key=key_text)
        audio_decoder = self.audio_decoder_config.random_init(key=key_audio)
        vocoder = self.vocoder_config.empty()
        return TTSModel(config=self, text_decoder=text_decoder, audio_decoder=audio_decoder, vocoder=vocoder)


class TTSModel(LalamoModule[TTSConfig]):
    text_decoder: TTSTextDecoder
    audio_decoder: TTSAudioDecoder
    vocoder: Vocoder

    @property
    def activation_precision(self) -> DTypeLike:
        return TTSConfig.activation_precision

    def export_weights(self) -> ParameterTree[Array]:
        return {}

    def import_weights(
        self,
        weights: ParameterTree[Array],
    ) -> Self:
        assert isinstance(weights, Mapping)
        return replace(
            self,
            text_decoder=self.text_decoder.import_weights(require_tree(weights["text_decoder"])),
            audio_decoder=self.audio_decoder.import_weights(require_tree(weights["audio_decoder"])),
        )



"""
Consts and configuration elements that were stored in Fishaudio
codebase as either global variables or magic consts
"""

DEFAULT_FISHAUDIO_RANDOM_SEED: int = 123
DEFAULT_FISH_AUDIO_SAMPLING_TEMPERATURE = 0.8008
DEFAULT_FISH_AUDIO_SAMPLING_TOP_P = 0.8008
DEFAULT_FISH_AUDIO_REPETITION_PENALTY: float = 1.1016
SHORT_LOGITS_SIZE: int = 1024
REPEAT_WINDOW_SIZE: int = 16
FISH_TIKTOKEN_PATTERN = "|".join(  # noqa: FLY002
    [
        r"(?i:'s|'t|'re|'ve|'m|'ll|'d)",
        r"\p{P}",
        r"[^\r\n\p{L}\p{N}]?\p{L}+",
        r"\p{N}",
        r" ?[^\s\p{L}\p{N}]+[\r\n]*",
        r"\s*[\r\n]+",
        r"\s+(\?!\S)",
        r"\s+",
    ],
)
IM_END_TOKEN = "<|im_end|>"


# NOTE: in fish-speech repo this was a YAML config stored right in the code
_default_audio_codec_config = {
    "_target_": "fish_speech.models.dac.modded_dac.DAC",
    "sample_rate": 44100,
    "encoder_dim": 64,
    "encoder_rates": [2, 4, 8, 8],
    "decoder_dim": 1536,
    "decoder_rates": [8, 8, 4, 2],
    "encoder_transformer_layers": [0, 0, 0, 4],
    "decoder_transformer_layers": [4, 0, 0, 0],
    "transformer_general_config": {
        "_target_": "fish_speech.models.dac.modded_dac.ModelArgs",
        "_partial_": True,
        "block_size": 16384,
        "n_local_heads": -1,
        "head_dim": 64,
        "rope_base": 10000,
        "norm_eps": 1e-5,
        "dropout_rate": 0.1,
        "attn_dropout_rate": 0.1,
        "channels_first": True,
    },
    "quantizer": {
        "_target_": "fish_speech.models.dac.rvq.DownsampleResidualVectorQuantize",
        "input_dim": 1024,
        "n_codebooks": 9,
        "codebook_size": 1024,
        "codebook_dim": 8,
        "quantizer_dropout": 0.5,
        "downsample_factor": [2, 2],
        "post_module": {
            "_target_": "fish_speech.models.dac.modded_dac.WindowLimitedTransformer",
            "causal": True,
            "window_size": 128,
            "input_dim": 1024,
            "config": {
                "_target_": "fish_speech.models.dac.modded_dac.ModelArgs",
                "block_size": 4096,
                "n_layer": 8,
                "n_head": 16,
                "dim": 1024,
                "intermediate_size": 3072,
                "n_local_heads": -1,
                "head_dim": 64,
                "rope_base": 10000,
                "norm_eps": 1e-5,
                "dropout_rate": 0.1,
                "attn_dropout_rate": 0.1,
                "channels_first": True,
            },
        },
        "pre_module": {
            "_target_": "fish_speech.models.dac.modded_dac.WindowLimitedTransformer",
            "causal": True,
            "window_size": 128,
            "input_dim": 1024,
            "config": {
                "_target_": "fish_speech.models.dac.modded_dac.ModelArgs",
                "block_size": 4096,
                "n_layer": 8,
                "n_head": 16,
                "dim": 1024,
                "intermediate_size": 3072,
                "n_local_heads": -1,
                "head_dim": 64,
                "rope_base": 10000,
                "norm_eps": 1e-5,
                "dropout_rate": 0.1,
                "attn_dropout_rate": 0.1,
                "channels_first": True,
            },
        },
        "semantic_codebook_size": 4096,
    },
}



from lalamo.model_import.model_configs import HFGPTOssConfig

from .common import ConfigMap, FileSpec, ModelSpec, WeightsType

__all__ = ["GPT_OSS_MODELS"]

GPT_OSS_MODELS = [
    ModelSpec(
        vendor="OpenAI",
        family="GPT-OSS",
        name="GPT-OSS-20B",
        size="20B",
        quantization=None,
        repo="openai/gpt-oss-20b",
        config_type=HFGPTOssConfig,
        weights_type=WeightsType.SAFETENSORS,
        configs=ConfigMap(
            chat_template=FileSpec("chat_template.jinja"),
        ),
    ),
]



from lalamo.model_import.model_configs import HFQwen3Config

from .common import ConfigMap, FileSpec, ModelSpec

__all__ = ["POLARIS_MODELS"]

POLARIS_MODELS = [
    ModelSpec(
        vendor="POLARIS-Project",
        family="Polaris-Preview",
        name="Polaris-4B-Preview",
        size="4B",
        quantization=None,
        repo="POLARIS-Project/Polaris-4B-Preview",
        config_type=HFQwen3Config,
        configs=ConfigMap(
            chat_template=FileSpec("chat_template.jinja"),
        ),
    ),
]



from typing import overload

import jax.numpy as jnp
import torch
from jaxtyping import Array

__all__ = ["jax_to_torch", "torch_to_jax"]

_typestr_to_torch_dtype = {
    "float16": torch.float16,
    "float32": torch.float32,
    "float64": torch.float64,
    "bfloat16": torch.bfloat16,
    "int8": torch.int8,
    "int16": torch.int16,
    "int32": torch.int32,
    "int64": torch.int64,
    "uint8": torch.uint8,
    "bool": torch.bool,
    "complex64": torch.complex64,
    "complex128": torch.complex128,
}
_torch_dtype_to_typestr = {v: k for k, v in _typestr_to_torch_dtype.items()}


def dtype_to_torch(in_type: str | jnp.dtype) -> torch.dtype:
    match in_type:
        case str():
            return _typestr_to_torch_dtype[in_type]
        case _:
            return _typestr_to_torch_dtype[jnp.dtype(in_type).name]


def dtype_to_jax(in_type: str | torch.dtype) -> jnp.dtype:
    match in_type:
        case str():
            return jnp.dtype(in_type)
        case torch.dtype():
            return jnp.dtype(_torch_dtype_to_typestr[in_type])


@torch.no_grad()
def _torch_to_jax_bfloat16(tensor: torch.Tensor) -> Array:
    if tensor.dtype != torch.bfloat16:
        raise ValueError("Trying to convert non-bfloat16 tensor to bfloat16")
    intermediate_tensor = tensor.view(torch.uint16)
    return jnp.array(intermediate_tensor).view("bfloat16")


@overload
def torch_to_jax(value: torch.Tensor) -> Array: ...
@overload
def torch_to_jax(value: str | torch.dtype) -> jnp.dtype: ...
def torch_to_jax(value: torch.Tensor | str | torch.dtype) -> Array | jnp.dtype:  # type: ignore[return-value]
    if isinstance(value, torch.Tensor):
        value = value.detach().cpu()
        if value.dtype == torch.bfloat16:
            return _torch_to_jax_bfloat16(value)
        return jnp.array(value.numpy())
    if isinstance(value, str | torch.dtype):
        return dtype_to_jax(value)


@overload
def jax_to_torch(value: Array) -> torch.Tensor: ...
@overload
def jax_to_torch(value: str | jnp.dtype) -> torch.dtype: ...
def jax_to_torch(value: Array | str | jnp.dtype) -> torch.Tensor | torch.dtype:
    if isinstance(value, str | jnp.dtype):
        return dtype_to_torch(value)
    if isinstance(value, Array):
        from torch.utils import dlpack as _dlpack

        if value.dtype == jnp.bfloat16:
            intermediate_array = value.view(jnp.uint16)
            return _dlpack.from_dlpack(intermediate_array).view(torch.bfloat16)
        return _dlpack.from_dlpack(value)



from .huggingface_message import load_hf_parquet, shuffle_dataset
from .utils import get_prefixes_ending_in_user_message

__all__ = [
    "get_prefixes_ending_in_user_message",
    "load_hf_parquet",
    "shuffle_dataset",
]



import json
from collections.abc import Mapping
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Literal, Self

from jaxtyping import DTypeLike

from lalamo.modules import (
    AttentionConfig,
    DecoderConfig,
    DeltaNetAttentionConfig,
    DenseMLPConfig,
    FullPrecisionLinearConfig,
    GroupQuantizedLinearConfig,
    MLXQuantizedTiedEmbeddingConfig,
    MLXQuantizedUntiedEmbeddingConfig,
    NormalizationConfig,
    TiedEmbeddingConfig,
    TransformerConfig,
    TransformerLayerConfig,
    UnscaledRoPEConfig,
    UntiedEmbeddingConfig,
    UpcastMode,
)
from lalamo.modules.activations import SiLU
from lalamo.modules.linear import MLXQuantizedLinearConfig
from lalamo.modules.token_mixers import SeparableCausalConvConfig
from lalamo.quantization import QuantizationMode

from .common import HuggingFaceLMConfig, MLXQuantizationConfig, QuantizationConfigType

__all__ = ["HFQwen35Config"]


@dataclass(frozen=True)
class HFQwen35Config(HuggingFaceLMConfig):
    model_type: Literal["qwen3_5"]
    hidden_size: int
    intermediate_size: int
    num_hidden_layers: int
    num_attention_heads: int
    num_key_value_heads: int
    max_position_embeddings: int
    rms_norm_eps: float
    tie_word_embeddings: bool
    vocab_size: int
    rope_parameters: dict[str, Any]

    head_dim: int
    attention_bias: bool
    layer_types: list[Literal["linear_attention", "full_attention"]]
    linear_conv_kernel_dim: int
    linear_key_head_dim: int
    linear_value_head_dim: int
    linear_num_key_heads: int
    linear_num_value_heads: int

    torch_dtype: Literal["bfloat16", "float16", "float32"] = "bfloat16"
    eos_token_id: int | list[int] = field(default_factory=list)
    quantization: QuantizationConfigType | None = None
    quantization_config: QuantizationConfigType | None = None

    @classmethod
    def from_json(cls, json_path: Path | str) -> Self:
        json_path = Path(json_path)
        with open(json_path) as f:
            config = json.load(f)
        # Multimodal configs nest text model params under text_config;
        # top-level keys override text_config on conflict.
        text_config = config.pop("text_config", {})
        merged = {**text_config, **config}
        return cls._converter.structure(merged, cls)

    def to_decoder_config(
        self,
        context_length: int | None,
        activation_precision: DTypeLike,
        accumulation_precision: DTypeLike,
        metadata_dict: Mapping[str, str],  # noqa: ARG002
    ) -> DecoderConfig:
        quantization = self.quantization or self.quantization_config

        is_mlx = isinstance(quantization, MLXQuantizationConfig)

        if is_mlx:
            if self.tie_word_embeddings:
                embedding_config = MLXQuantizedTiedEmbeddingConfig(
                    input_scale=None,
                    logit_soft_cap=None,
                    group_size=quantization.group_size,
                    embedding_quantization_mode=QuantizationMode.from_num_bits(quantization.bits),
                    activation_quantization_mode=None,
                    activation_precision=activation_precision,
                )
            else:
                embedding_config = MLXQuantizedUntiedEmbeddingConfig(
                    input_scale=None,
                    logit_soft_cap=None,
                    group_size=quantization.group_size,
                    embedding_quantization_mode=QuantizationMode.from_num_bits(quantization.bits),
                    activation_quantization_mode=None,
                    activation_precision=activation_precision,
                )
        else:  # noqa: PLR5501
            if self.tie_word_embeddings:
                embedding_config = TiedEmbeddingConfig(
                    input_scale=None,
                    logit_soft_cap=None,
                    precision=activation_precision,
                )
            else:
                embedding_config = UntiedEmbeddingConfig(
                    input_scale=None,
                    logit_soft_cap=None,
                    precision=activation_precision,
                )

        rope_config = UnscaledRoPEConfig(
            precision=activation_precision,
            base=self.rope_parameters["rope_theta"],
            max_sequence_length=context_length or self.max_position_embeddings,
        )

        # Qwen3.5 RMSNorm computes (1 + weight) * norm(x). HF stores raw weights,
        # but MLX community converters bake the +1 into the weights, so no offset needed.
        rmsnorm_config = NormalizationConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.rms_norm_eps,
            scale_offset=None if is_mlx else 1.0,
            upcast_mode=UpcastMode.ONLY_NORMALIZATION,
            subtract_mean=False,
        )
        # Qwen3.5 DeltaNet RMSNorm uses direct weights (no +1 offset).
        gated_rmsnorm_config = NormalizationConfig(
            scale_precision=activation_precision,
            accumulation_precision=accumulation_precision,
            epsilon=self.rms_norm_eps,
            scale_offset=None,
            upcast_mode=UpcastMode.ONLY_NORMALIZATION,
            subtract_mean=False,
        )

        if quantization is None:
            linear_config = FullPrecisionLinearConfig(
                precision=activation_precision,
            )
        elif isinstance(quantization, MLXQuantizationConfig):
            linear_config = MLXQuantizedLinearConfig(
                group_size=quantization.group_size,
                weight_quantization_mode=QuantizationMode.from_num_bits(quantization.bits),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )
        else:
            linear_config = GroupQuantizedLinearConfig(
                group_size=quantization.group_size,
                weight_quantization_mode=QuantizationMode.from_num_bits(quantization.bits),
                activation_quantization_mode=None,
                activation_precision=activation_precision,
            )

        partial_rotary_factor = self.rope_parameters["partial_rotary_factor"]

        layer_configs = []
        for layer_type in self.layer_types:
            if layer_type == "linear_attention":
                mixer_config = DeltaNetAttentionConfig(
                    in_proj_config=linear_config,
                    conv_config=SeparableCausalConvConfig(
                        precision=activation_precision,
                        has_biases=False,
                    ),
                    out_proj_config=linear_config,
                    norm_config=gated_rmsnorm_config,
                    num_heads=self.linear_num_value_heads,
                    num_groups=self.linear_num_key_heads,
                    head_dim=self.linear_key_head_dim,
                    value_head_dim=self.linear_value_head_dim,
                    kernel_size=self.linear_conv_kernel_dim,
                )
            else:
                mixer_config = AttentionConfig(
                    qkv_projection_config=linear_config,
                    out_projection_config=linear_config,
                    query_norm_config=rmsnorm_config,
                    key_norm_config=rmsnorm_config,
                    logit_soft_cap=None,
                    has_sinks=False,
                    has_qkv_biases=self.attention_bias,
                    has_out_biases=self.attention_bias,
                    num_heads=self.num_attention_heads,
                    num_groups=self.num_key_value_heads,
                    head_dim=self.head_dim,
                    is_causal=True,
                    scale=None,
                    sliding_window_size=None,
                    gate_projection_config=linear_config,
                    partial_rope_dim=int(self.head_dim * partial_rotary_factor),
                )

            layer_configs.append(
                TransformerLayerConfig(
                    pre_mixer_norm_config=rmsnorm_config,
                    mixer_config=mixer_config,
                    post_mixer_norm_config=None,
                    pre_mlp_norm_config=rmsnorm_config,
                    mlp_config=DenseMLPConfig(
                        linear_config=linear_config,
                        activation=SiLU(),
                        has_up_biases=False,
                        has_down_biases=False,
                        up_clipping=None,
                        gate_clipping=None,
                    ),
                    post_mlp_norm_config=None,
                ),
            )

        transformer_config = TransformerConfig(
            global_rope_config=rope_config,
            local_rope_config=None,
            layer_configs=tuple(layer_configs),
            output_norm_config=rmsnorm_config,
            model_dim=self.hidden_size,
            hidden_dim=self.intermediate_size,
            context_length=context_length or self.max_position_embeddings,
        )

        return DecoderConfig(
            embedding_config=embedding_config,
            transformer_config=transformer_config,
            vocab_size=self.vocab_size,
        )



import math
from collections.abc import Mapping
from dataclasses import dataclass, replace
from typing import Self

import equinox as eqx
import jax
import jax.numpy as jnp
from jax import lax
from jaxtyping import Array, DTypeLike, Float, PRNGKeyArray

from lalamo.common import ParameterTree, dummy_array
from lalamo.modules.common import LalamoModule


def _get_extra_padding_for_conv1d(length: int, kernel_size: int, stride: int, padding_total: int = 0) -> int:
    """Calculate extra padding needed to ensure output length is correct."""
    n_frames = (length - kernel_size + padding_total) / stride + 1
    ideal_length = (math.ceil(n_frames) - 1) * stride + (kernel_size - padding_total)
    return int(ideal_length - length)


@dataclass(frozen=True)
class CausalConv1dConfig:
    precision: DTypeLike
    has_biases: bool

    def random_init(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        dilation: int = 1,
        groups: int = 1,
        *,
        key: PRNGKeyArray,
    ) -> "CausalConv1d":
        effective_kernel_size = (kernel_size - 1) * dilation + 1
        weights = jax.random.normal(
            key,
            shape=(out_channels, in_channels // groups, kernel_size),
            dtype=self.precision,
        )

        if self.has_biases:
            biases = jnp.zeros((out_channels,), dtype=self.precision)
        else:
            biases = None

        return CausalConv1d(
            config=self,
            weights=weights,
            biases=biases,
            stride=stride,
            dilation=dilation,
            groups=groups,
            effective_kernel_size=effective_kernel_size,
        )

    def empty(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        dilation: int = 1,
        groups: int = 1,
    ) -> "CausalConv1d":
        effective_kernel_size = (kernel_size - 1) * dilation + 1

        weights = dummy_array(
            (out_channels, in_channels // groups, kernel_size),
            dtype=self.precision,
        )

        if self.has_biases:
            biases = dummy_array((out_channels,), dtype=self.precision)
        else:
            biases = None

        return CausalConv1d(
            config=self,
            weights=weights,
            biases=biases,
            stride=stride,
            dilation=dilation,
            groups=groups,
            effective_kernel_size=effective_kernel_size,
        )


class CausalConv1d(LalamoModule[CausalConv1dConfig]):
    """Causal 1D convolution module.
    Implements causal convolution by left-padding the input with zeros.
    The output at position t only depends on inputs at positions <= t.
    """

    weights: Float[Array, "out_channels in_channels_per_group kernel_size"]
    biases: Float[Array, " out_channels"] | None

    stride: int = eqx.field(static=True)
    dilation: int = eqx.field(static=True)
    groups: int = eqx.field(static=True)
    effective_kernel_size: int = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def out_channels(self) -> int:
        out_channels, _, _ = self.weights.shape
        return out_channels

    @property
    def in_channels(self) -> int:
        _, in_channels_per_group, _ = self.weights.shape
        return in_channels_per_group * self.groups

    @property
    def kernel_size(self) -> int:
        _, _, kernel_size = self.weights.shape
        return kernel_size

    @property
    def padding(self) -> int:
        return self.effective_kernel_size - self.stride

    def __call__(
        self,
        x: Float[Array, "batch sequence in_channels"],
    ) -> Float[Array, "batch sequence_out out_channels"]:
        length = x.shape[1]  # sequence dimension is axis 1
        pad = self.padding
        extra_padding = _get_extra_padding_for_conv1d(length, self.effective_kernel_size, self.stride, pad)

        # Pad on the left (causal padding) and extra on the right if needed
        # Input is (batch, sequence, channels), pad the sequence dimension (axis 1)
        x_padded = jnp.pad(x, ((0, 0), (pad, extra_padding), (0, 0)), mode="constant", constant_values=0)

        # Input: (N, S, C) - NSC format
        # Kernel: (O, I, K) - OIH format
        # Output: (N, S, C) - NSC format
        output = lax.conv_general_dilated(
            x_padded,
            self.weights,
            window_strides=(self.stride,),
            padding="VALID",
            rhs_dilation=(self.dilation,),
            feature_group_count=self.groups,
            dimension_numbers=("NHC", "OIH", "NHC"),
        )

        if self.biases is not None:
            output = output + self.biases[None, None, :]

        return output

    def export_weights(self) -> ParameterTree[Array]:
        result: dict[str, Array] = {"weights": self.weights}
        if self.biases is not None:
            result["biases"] = self.biases
        return result

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["weights"], Array)
        if self.biases is not None:
            assert isinstance(weights["biases"], Array)
            biases = weights["biases"]
        else:
            biases = None
        return replace(
            self,
            weights=weights["weights"],
            biases=biases,
        )


@dataclass(frozen=True)
class CausalTransposeConv1dConfig:
    precision: DTypeLike
    has_biases: bool

    def random_init(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        groups: int = 1,
        *,
        key: PRNGKeyArray,
    ) -> "CausalTransposeConv1d":
        # Weight shape: (out_channels, in_channels // groups, kernel_size) - JAX OIK format
        in_per_group = in_channels // groups
        weights = jax.random.normal(
            key,
            (out_channels, in_per_group, kernel_size),
            dtype=self.precision,
        )

        if self.has_biases:
            biases = jnp.zeros((out_channels,), dtype=self.precision)
        else:
            biases = None

        return CausalTransposeConv1d(
            config=self,
            weights=weights,
            biases=biases,
            in_channels=in_channels,
            stride=stride,
            groups=groups,
        )

    def empty(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: int,
        stride: int = 1,
        groups: int = 1,
    ) -> "CausalTransposeConv1d":
        # Weight shape: (out_channels, in_channels // groups, kernel_size) - JAX OIK format
        in_per_group = in_channels // groups
        weights = dummy_array(
            (out_channels, in_per_group, kernel_size),
            dtype=self.precision,
        )

        if self.has_biases:
            biases = dummy_array((out_channels,), dtype=self.precision)
        else:
            biases = None

        return CausalTransposeConv1d(
            config=self,
            weights=weights,
            biases=biases,
            in_channels=in_channels,
            stride=stride,
            groups=groups,
        )


class CausalTransposeConv1d(LalamoModule[CausalTransposeConv1dConfig]):
    """Causal transposed 1D convolution (deconvolution) with groups support.

    Implements causal transposed convolution by removing appropriate padding
    from the output. Supports grouped convolutions for efficient upsampling.

    Weight format: (out_channels, in_channels // groups, kernel_size) - JAX OIK format
    with kernel already flipped for transposed convolution.
    """

    weights: Float[Array, "out_channels in_channels_per_group kernel_size"]
    biases: Float[Array, " out_channels"] | None

    in_channels: int = eqx.field(static=True)
    stride: int = eqx.field(static=True)
    groups: int = eqx.field(static=True)

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def out_channels(self) -> int:
        out_channels, _, _ = self.weights.shape
        return out_channels

    @property
    def kernel_size(self) -> int:
        _, _, kernel_size = self.weights.shape
        return kernel_size

    def __call__(
        self,
        x: Float[Array, "batch sequence in_channels"],
    ) -> Float[Array, "batch sequence_out out_channels"]:
        # Input: (N, S, C) - NSC format
        # Kernel: (O, I/g, K) - JAX OIK format, already transformed and flipped
        # Output: (N, S, C) - NSC format

        # Calculate padding for transposed convolution
        # Output length = (input_length - 1) * stride + kernel_size
        # We want VALID-like behavior then trim
        padding_needed = self.kernel_size - 1
        padding = ((padding_needed, padding_needed),)

        output = lax.conv_general_dilated(
            x,
            self.weights,
            window_strides=(1,),
            padding=padding,
            lhs_dilation=(self.stride,),
            rhs_dilation=(1,),
            dimension_numbers=("NHC", "OIH", "NHC"),
            feature_group_count=self.groups,
        )

        if self.biases is not None:
            output = output + self.biases[None, None, :]

        # Remove padding to maintain causality
        # NanoCodec uses trim_right_ratio=1, meaning all padding trimmed from right
        pad = self.kernel_size - self.stride
        padding_right = math.ceil(pad)
        padding_left = pad - padding_right

        # Unpad: remove padding_left from start and padding_right from end
        if padding_left > 0 or padding_right > 0:
            end = output.shape[1] - padding_right if padding_right > 0 else output.shape[1]
            output = output[:, padding_left:end, :]

        return output

    def export_weights(self) -> ParameterTree[Array]:
        result: dict[str, Array] = {"weights": self.weights}
        if self.biases is not None:
            result["biases"] = self.biases
        return result

    def import_weights(self, weights: ParameterTree[Array]) -> Self:
        assert isinstance(weights, Mapping)
        assert isinstance(weights["weights"], Array)
        if self.biases is not None:
            assert isinstance(weights["biases"], Array)
            biases = weights["biases"]
        else:
            biases = None
        return replace(
            self,
            weights=weights["weights"],
            biases=biases,
        )


@dataclass(frozen=True)
class Snake1dConfig:
    precision: DTypeLike

    def empty(self, channels: int) -> "Snake1d":
        alpha = dummy_array((channels,), dtype=self.precision)
        return Snake1d(config=self, alpha=alpha)

    def random_init(self, channels: int) -> "Snake1d":
        alpha = jnp.ones((channels,), dtype=self.precision)
        return Snake1d(config=self, alpha=alpha)


class Snake1d(LalamoModule[Snake1dConfig]):
    """Snake1d activation module.
    Implements the Snake activation function: x + (1/alpha) * sin^2(alpha * x)
    """

    alpha: Float[Array, " channels"]

    @property
    def activation_precision(self) -> DTypeLike:
        return self.config.precision

    @property
    def channels(self) -> int:
        return self.alpha.shape[0]

    def __call__(
        self,
        x: Float[Array, "batch sequence channels"],
    ) -> Float[Array, "batch sequence channels"]:
        # alpha is (channels,), broadcast to (1, 1, channels) for (batch, seq, channels) input
        alpha = self.alpha[None, None, :]
        # Snake activation: x + (1/alpha) * sin^2(alpha * x)
        return x + jnp.reciprocal(alpha + 1e-9) * jnp.square(jnp.sin(alpha * x))

    def export_weights(self) -> ParameterTree[Array]:
        return {"alpha": self.alpha}

    def import_weights(self, weights: ParameterTree[Array]) -> "Snake1d":
        assert isinstance(weights, Mapping)
        assert isinstance(weights["alpha"], Array)
        return replace(self, alpha=weights["alpha"])



import importlib.metadata
import json
import tarfile
import tempfile
from collections import ChainMap
from collections.abc import Callable, Iterator
from contextlib import ExitStack, contextmanager
from dataclasses import dataclass, replace
from pathlib import Path
from tarfile import TarInfo
from typing import NamedTuple

import huggingface_hub
import jax.numpy as jnp
import yaml
from jax import Array
from jaxtyping import DTypeLike
from tokenizers import Tokenizer

from lalamo.audio.tts_message_processor import TTSMessageProcessor, TTSMessageProcessorConfig
from lalamo.audio.utils import dummy_char_level_tokenizer_config
from lalamo.message_processor import MessageProcessor, MessageProcessorConfig
from lalamo.model_import.model_configs.huggingface.fishaudio import FishAudioConfig
from lalamo.model_registry import ModelRegistry
from lalamo.models import (
    ClassifierModel,
    ClassifierModelConfig,
    GenerationConfig,
    LanguageModel,
    LanguageModelConfig,
    TTSGenerator,
    TTSGeneratorConfig,
)
from lalamo.modules import Classifier, Decoder, LalamoModule, TTSModel
from lalamo.modules.common import ShardingConfig, use_sharding
from lalamo.quantization import QuantizationMode
from lalamo.utils import process_chat_template

from .huggingface_generation_config import HFGenerationConfig, _policy_from_hf_config, merge_token_ids
from .huggingface_tokenizer_config import HFTokenizerConfig
from .model_configs import ForeignClassifierConfig, ForeignConfig, ForeignLMConfig
from .model_specs import FileSpec, ModelSpec, ModelType, UseCase
from .model_specs.common import JSONFieldSpec, WeightsType

__all__ = [
    "DownloadingFileEvent",
    "FinishedDownloadingFileEvent",
    "InitializingModelEvent",
    "ModelMetadata",
    "ModelSpec",
    "ModelType",
    "StatusEvent",
    "download_file",
    "import_model",
]


LALAMO_VERSION = importlib.metadata.version("lalamo")


class DownloadingFileEvent(NamedTuple):
    file: FileSpec


class FinishedDownloadingFileEvent(NamedTuple):
    file: FileSpec


class InitializingModelEvent(NamedTuple):
    pass


class FinishedInitializingModelEvent(NamedTuple):
    pass


type StatusEvent = (
    DownloadingFileEvent | FinishedDownloadingFileEvent | InitializingModelEvent | FinishedInitializingModelEvent
)


@dataclass(frozen=True)
class ModelMetadata:
    toolchain_version: str
    vendor: str
    family: str
    name: str
    size: str
    quantization: QuantizationMode | None
    repo: str
    use_cases: tuple[UseCase, ...]
    model_type: ModelType
    model_config: LanguageModelConfig | ClassifierModelConfig | TTSGeneratorConfig
    grammar_start_tokens: tuple[str, ...]


def download_file(
    file_spec: FileSpec,
    model_repo: str,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> Path:
    if progress_callback is not None:
        progress_callback(DownloadingFileEvent(file_spec))
    result = huggingface_hub.hf_hub_download(
        repo_id=file_spec.repo or model_repo,
        local_dir=output_dir,
        filename=file_spec.filename,
    )
    if progress_callback is not None:
        progress_callback(FinishedDownloadingFileEvent(file_spec))
    return Path(result)


def list_weight_files(model_repo: str, weights_type: WeightsType) -> list[FileSpec]:
    all_files = huggingface_hub.list_repo_files(model_repo)
    match weights_type:
        case WeightsType.SAFETENSORS:
            return [FileSpec(filename) for filename in all_files if filename.endswith(".safetensors")]
        case WeightsType.TORCH:
            return [FileSpec(filename) for filename in all_files if filename.endswith(".pth")]
        case WeightsType.NEMO:
            return [FileSpec(filename) for filename in all_files if filename.endswith(".nemo")]


def download_weights(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> list[Path]:
    return [
        download_file(file_spec, model_spec.repo, output_dir, progress_callback)
        for file_spec in list_weight_files(model_spec.repo, model_spec.weights_type)
    ]


def download_config_file(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> Path:
    return download_file(model_spec.configs.model_config, model_spec.repo, output_dir, progress_callback)


class ImportResults(NamedTuple):
    model: LanguageModel | ClassifierModel | TTSGenerator
    metadata: ModelMetadata


def token_ids_to_text(tokenizer: Tokenizer, token_ids: int | list[int] | None) -> str | None:
    if isinstance(token_ids, int):
        token_ids = [token_ids]

    if not isinstance(token_ids, list) or any((not isinstance(el, int)) for el in token_ids):
        return None

    decoded = tokenizer.decode(token_ids[:1], skip_special_tokens=False)
    return decoded


def _instantiate_tokenizer_from_model_spec(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> Tokenizer:
    if model_spec.vendor == "NVIDIA" and model_spec.family == "nanocodec":
        # NOTE: once text decoder for Nanocodec is implemented - proper Tokenizer will hopefully become available
        tokenizer = Tokenizer.from_str(dummy_char_level_tokenizer_config())
    else:
        assert isinstance(model_spec.configs.tokenizer, FileSpec)
        tokenizer_file = download_file(model_spec.configs.tokenizer, model_spec.repo, output_dir, progress_callback)
        tokenizer = Tokenizer.from_file(str(tokenizer_file))
    return tokenizer


def import_message_processor(
    model_spec: ModelSpec,
    output_dir: Path | str | None = None,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> MessageProcessor:
    tokenizer_config_file = download_file(
        model_spec.configs.tokenizer_config,
        model_spec.repo,
        output_dir,
        progress_callback,
    )
    tokenizer_config = HFTokenizerConfig.from_json(tokenizer_config_file)

    if tokenizer_config.chat_template is None:
        match model_spec.configs.chat_template:
            case JSONFieldSpec(file_spec, field_name):
                json_file = download_file(file_spec, model_spec.repo, output_dir)
                with open(json_file) as file:
                    json_dict = json.load(file)
                prompt_template = json_dict[field_name]
            case FileSpec(_) as file_spec:
                chat_template_file = download_file(file_spec, model_spec.repo, output_dir)
                prompt_template = chat_template_file.read_text()
            case str() as template_string:
                prompt_template = template_string
            case None:
                raise ValueError("No chat template specified.")
    else:
        if model_spec.configs.chat_template is not None:
            raise ValueError("Conflicting chat template specifications.")
        prompt_template = tokenizer_config.chat_template

    prompt_template = process_chat_template(prompt_template)
    tokenizer = _instantiate_tokenizer_from_model_spec(model_spec, output_dir, progress_callback)

    added_tokens = tokenizer_config.added_tokens()
    added_special_tokens = [token for token in added_tokens if token.special]
    added_not_special_tokens = [token for token in added_tokens if not token.special]
    tokenizer.add_special_tokens(added_special_tokens)
    tokenizer.add_tokens(added_not_special_tokens)

    bos_token = getattr(tokenizer_config, "bos_token", None)
    eos_token = getattr(tokenizer_config, "eos_token", None)

    # If we were not able to identify bos/eos - they are probably somewhere else, so we check config.json
    if eos_token is None or bos_token is None:
        foreign_decoder_config_file = download_config_file(model_spec, output_dir, progress_callback)
        with open(foreign_decoder_config_file) as foreign_decoder_file:
            foreign_decoder_json = json.load(foreign_decoder_file)

        if bos_token is None:
            bos_token_id: int | list[int] | None = foreign_decoder_json.get("bos_token_id")
            bos_token = token_ids_to_text(tokenizer, bos_token_id)
        if eos_token is None:
            eos_token_id: int | list[int] | None = foreign_decoder_json.get("eos_token_id")
            eos_token = token_ids_to_text(tokenizer, eos_token_id)

    system_prompt_text = None
    match model_spec.configs.system_prompt:
        case FileSpec(_) as file_spec:
            system_prompt_file = download_file(file_spec, model_spec.repo, output_dir, progress_callback)
            system_prompt_text = system_prompt_file.read_text()
        case str() as sp:
            system_prompt_text = sp
        case None:
            pass

    message_processor_config = MessageProcessorConfig(
        prompt_template=prompt_template,
        output_parser_regex=model_spec.output_parser_regex,
        system_role_name=model_spec.system_role_name,
        user_role_name=model_spec.user_role_name,
        assistant_role_name=model_spec.assistant_role_name,
        bos_token=bos_token,
        eos_token=eos_token,
        default_system_prompt=system_prompt_text,
    )
    return MessageProcessor(config=message_processor_config, tokenizer=tokenizer)


@contextmanager
def _unpack_nemo_model(nemo_model_path: Path) -> Iterator[tuple[list[Path], Path]]:
    def _is_safe_to_extract(tar_item_info: TarInfo) -> bool:
        return not (
            tar_item_info.name.startswith("..") or Path(tar_item_info.name).is_absolute() or tar_item_info.size == 0
        )

    with tempfile.TemporaryDirectory() as tmpdir:
        # NOTE: checking each member of tar archive to avoid potential
        # attack by extracting file to arbitrary location
        with tarfile.open(nemo_model_path, "r") as tar:
            for tar_member in tar.getmembers():
                if _is_safe_to_extract(tar_member):
                    tar.extract(tar_member.name, path=tmpdir)

        # go through files in extracted model and locate Torch model file and model YAML config file
        weights_paths = list(Path(tmpdir).glob("*.ckpt"))
        if not weights_paths:
            raise FileNotFoundError("Failed to find Nemo model weights")
        (yaml_config_path,) = list(Path(tmpdir).glob("*.yaml"))

        # load YAML config and re-save it in JSON format
        with open(yaml_config_path) as f:
            config_yaml = yaml.safe_load(f)
        config_path = yaml_config_path.with_suffix(".json")
        with open(config_path, "w") as f:
            json.dump(config_yaml, f)

        yield (weights_paths, config_path)


@contextmanager
def _download_weights_and_config_files(
    model_spec: ModelSpec,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> Iterator[tuple[list[Path], Path]]:
    if model_spec.weights_type == WeightsType.NEMO:
        (nemo_model_file,) = download_weights(model_spec, progress_callback=progress_callback)
        with _unpack_nemo_model(nemo_model_file) as nemo_file_contents:
            weights_paths, foreign_config_file_path = nemo_file_contents
            yield (weights_paths, foreign_config_file_path)
    else:
        weights_paths = download_weights(model_spec, progress_callback=progress_callback)
        foreign_config_file_path = download_config_file(model_spec)

        yield (weights_paths, foreign_config_file_path)


def _load_main_processing_module(
    model_spec: ModelSpec,
    weights_paths: list[Path],
    precision: DTypeLike,
    foreign_config: ForeignConfig,
    progress_callback: Callable[[StatusEvent], None] | None = None,
    context_length: int | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
) -> LalamoModule:
    with ExitStack() as stack:
        weights_shards = []
        metadata_shards = []
        for weights_path in weights_paths:
            weights_shard, metadata_shard = stack.enter_context(model_spec.weights_type.load(weights_path, precision))
            weights_shards.append(weights_shard)
            metadata_shards.append(metadata_shard)
        weights_dict: ChainMap[str, Array] = ChainMap(*weights_shards)
        metadata_dict: ChainMap[str, str] = ChainMap(*metadata_shards)

        if progress_callback is not None:
            progress_callback(InitializingModelEvent())

        processing_module = foreign_config.load(
            context_length,
            precision,
            accumulation_precision,
            weights_dict,
            metadata_dict,
        )

    return processing_module


def _import_language_model(
    model_spec: ModelSpec,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> tuple[LanguageModel, LanguageModelConfig]:
    with _download_weights_and_config_files(
        model_spec,
        progress_callback=progress_callback,
    ) as (model_weights_paths, config_path):
        foreign_decoder_config = model_spec.config_type.from_json(config_path)
        assert isinstance(foreign_decoder_config, ForeignLMConfig)

        if precision is None:
            precision = foreign_decoder_config.default_precision
        decoder = _load_main_processing_module(
            model_spec,
            model_weights_paths,
            precision,
            foreign_decoder_config,
            progress_callback,
            context_length,
            accumulation_precision,
        )
        assert isinstance(decoder, Decoder)

    if progress_callback is not None:
        progress_callback(FinishedInitializingModelEvent())

    message_processor = import_message_processor(model_spec)

    stop_token_ids = merge_token_ids(foreign_decoder_config.eos_token_ids)

    if isinstance(model_spec.configs.generation_config, GenerationConfig):
        candidate_generation_config = model_spec.configs.generation_config
        stop_token_ids = merge_token_ids(candidate_generation_config.stop_token_ids, stop_token_ids)
        generation_config = replace(candidate_generation_config, stop_token_ids=stop_token_ids)
    elif isinstance(model_spec.configs.generation_config, FileSpec):
        hf_generation_config_file = download_file(model_spec.configs.generation_config, model_spec.repo)
        hf_generation_config = HFGenerationConfig.from_json(hf_generation_config_file)
        stop_token_ids = merge_token_ids(stop_token_ids, hf_generation_config.eos_token_id)
        generation_config = _policy_from_hf_config(hf_generation_config, stop_token_ids=stop_token_ids)
    else:
        generation_config = GenerationConfig(stop_token_ids)

    language_model_config = LanguageModelConfig(
        model_config=decoder.config,
        message_processor_config=message_processor.config,
        generation_config=generation_config,
    )

    language_model = LanguageModel(language_model_config, decoder, message_processor)
    return language_model, language_model_config


def _import_classifier(
    model_spec: ModelSpec,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> tuple[ClassifierModel, ClassifierModelConfig]:
    with _download_weights_and_config_files(
        model_spec,
        progress_callback=progress_callback,
    ) as (model_weights_paths, config_path):
        foreign_classifier_config = model_spec.config_type.from_json(config_path)
        assert isinstance(foreign_classifier_config, ForeignClassifierConfig)

        if precision is None:
            precision = foreign_classifier_config.default_precision

        classifier = _load_main_processing_module(
            model_spec,
            model_weights_paths,
            precision,
            foreign_classifier_config,
            progress_callback,
            context_length,
            accumulation_precision,
        )
        assert isinstance(classifier, Classifier)

    if progress_callback is not None:
        progress_callback(FinishedInitializingModelEvent())

    message_processor = import_message_processor(model_spec)

    classifier_model_config = ClassifierModelConfig(
        model_config=classifier.config,
        message_processor_config=message_processor.config,
    )
    classifier_model = ClassifierModel(classifier_model_config, classifier, message_processor)
    return classifier_model, classifier_model_config


def _import_tts_model(
    model_spec: ModelSpec,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
) -> tuple[TTSGenerator, TTSGeneratorConfig]:
    with _download_weights_and_config_files(
        model_spec,
        progress_callback=progress_callback,
    ) as (model_weights_paths, config_path):
        foreign_tts_config = model_spec.config_type.from_json(config_path)
        if precision is None:
            precision = foreign_tts_config.default_precision
        if model_spec.vendor == "FishAudio" and model_spec.family == "openaudio":
            # NOTE: for FishAudio model we need certain info from Tokenizer even during inference stage
            # so we load the Tokenizer and update config using data from it
            from lalamo.model_import.loaders.fishaudio_loaders import load_tokenizer_from_fishaudio_tiktoken

            assert isinstance(model_spec.configs.tokenizer, FileSpec)
            tokenizer_path = download_file(model_spec.configs.tokenizer, model_repo=model_spec.repo)

            tokenizer_special_tokens_path = download_file(
                FileSpec(filename="special_tokens.json"),
                model_repo=model_spec.repo,
            )
            tokenizer, special_inference_tokens = load_tokenizer_from_fishaudio_tiktoken(
                tokenizer_path,
                tokenizer_special_tokens_path,
            )
            assert isinstance(foreign_tts_config, FishAudioConfig)
            foreign_tts_config = replace(
                foreign_tts_config,
                semantic_token_begin_id=special_inference_tokens.semantic_begin_id,
                semantic_token_end_id=special_inference_tokens.semantic_end_id,
                im_end_token_id=special_inference_tokens.im_end_token_id,
            )
        else:
            tokenizer = _instantiate_tokenizer_from_model_spec(model_spec, None, progress_callback)

        tts_model = _load_main_processing_module(
            model_spec,
            model_weights_paths,
            precision,
            foreign_tts_config,
            progress_callback,
            context_length,
            accumulation_precision,
        )

        assert isinstance(tts_model, TTSModel)
        if progress_callback is not None:
            progress_callback(FinishedInitializingModelEvent())

    assert isinstance(model_spec.configs.chat_template, str)
    tts_request_factory_config = TTSMessageProcessorConfig(
        prompt_template=model_spec.configs.chat_template,
    )
    message_processor = TTSMessageProcessor(tts_request_factory_config, tokenizer)

    tts_generator_config = TTSGeneratorConfig(
        tts_config=foreign_tts_config.to_lalamo_config(
            context_length=context_length,
            activation_precision=precision,
            accumulation_precision=accumulation_precision,
            metadata_dict={},
        ),
        message_processor_config=message_processor.config,
    )
    tts_generator = TTSGenerator(tts_generator_config, tts_model, message_processor)

    return (tts_generator, tts_generator_config)


def import_model(
    model_spec: ModelSpec | str,
    *,
    context_length: int | None = None,
    precision: DTypeLike | None = None,
    accumulation_precision: DTypeLike = jnp.float32,
    progress_callback: Callable[[StatusEvent], None] | None = None,
    sharding_config: ShardingConfig | None = None,
) -> ImportResults:
    if isinstance(model_spec, str):
        try:
            model_spec = ModelRegistry.build().repo_to_model[model_spec]
        except KeyError as e:
            raise ValueError(f"Unknown model: {model_spec}") from e

    with use_sharding(sharding_config):
        match model_spec.model_type:
            case ModelType.LANGUAGE_MODEL:
                model, config = _import_language_model(
                    model_spec,
                    context_length=context_length,
                    precision=precision,
                    accumulation_precision=accumulation_precision,
                    progress_callback=progress_callback,
                )
            case ModelType.CLASSIFIER_MODEL:
                model, config = _import_classifier(
                    model_spec,
                    context_length=context_length,
                    precision=precision,
                    accumulation_precision=accumulation_precision,
                    progress_callback=progress_callback,
                )
            case ModelType.TTS_MODEL:
                model, config = _import_tts_model(
                    model_spec,
                    context_length=context_length,
                    precision=precision,
                    accumulation_precision=accumulation_precision,
                    progress_callback=progress_callback,
                )

    metadata = ModelMetadata(
        toolchain_version=LALAMO_VERSION,
        vendor=model_spec.vendor,
        family=model_spec.family,
        name=model_spec.name,
        size=model_spec.size,
        quantization=model_spec.quantization,
        repo=model_spec.repo,
        use_cases=model_spec.use_cases,
        model_type=model_spec.model_type,
        model_config=config,
        grammar_start_tokens=model_spec.grammar_start_tokens,
    )
    return ImportResults(model, metadata)



from collections.abc import (
    Callable,
    Iterator,
    Mapping,
)
from contextlib import contextmanager
from dataclasses import dataclass, field
from enum import Enum, StrEnum
from pathlib import Path
from typing import Any, ClassVar, cast, get_args, get_origin

import cattrs
import jax.numpy as jnp
from jaxtyping import DTypeLike

from lalamo.common import cast_if_float
from lalamo.model_import.model_configs import ForeignConfig
from lalamo.models.language_model import GenerationConfig
from lalamo.quantization import QuantizationMode
from lalamo.safetensors import safe_read
from lalamo.utils import MapDictValues

__all__ = [
    "ConfigMap",
    "FileSpec",
    "JSONFieldSpec",
    "ModelSpec",
    "ModelType",
    "UseCase",
    "WeightsType",
    "awq_model_spec",
    "build_quantized_models",
]


class ModelType(StrEnum):
    LANGUAGE_MODEL = "language_model"
    CLASSIFIER_MODEL = "classifier_model"
    TTS_MODEL = "tts_model"


class WeightsType(Enum):
    SAFETENSORS = "safetensors"
    TORCH = "torch"
    NEMO = "nemo"

    @contextmanager
    def load(
        self,
        filename: Path | str,
        float_dtype: DTypeLike,
    ) -> Iterator[tuple[Mapping[str, jnp.ndarray], Mapping[str, str]]]:
        if self == WeightsType.SAFETENSORS:
            with Path(filename).open("rb") as fd:
                (metadata_dict, weights_dict) = safe_read(fd)
                yield MapDictValues(lambda v: cast_if_float(v, float_dtype), weights_dict), metadata_dict or {}
        else:
            # NOTE: this path also includes Nemo models, because we expect the model to be downloaded and
            # extracted. After that toch checkpoint contained in the model will later be processed here.
            import torch

            from lalamo.modules.torch_interop import torch_to_jax

            torch_weights = torch.load(filename, map_location="cpu", weights_only=True)
            yield MapDictValues(lambda v: cast_if_float(torch_to_jax(v), float_dtype), torch_weights), {}


class UseCase(Enum):
    CODE = "code"


@dataclass(frozen=True)
class FileSpec:
    filename: str
    repo: str | None = None


@dataclass(frozen=True)
class JSONFieldSpec:
    file_spec: FileSpec
    field_name: str


@dataclass(frozen=True)
class ConfigMap:
    model_config: FileSpec = field(default=FileSpec("config.json"))
    tokenizer: FileSpec | None = field(default=FileSpec("tokenizer.json"))
    tokenizer_config: FileSpec = field(default=FileSpec("tokenizer_config.json"))
    generation_config: FileSpec | GenerationConfig | None = field(default=FileSpec("generation_config.json"))
    chat_template: FileSpec | JSONFieldSpec | str | None = None
    system_prompt: FileSpec | str | None = None


def _is_foreign_config_type(t: object) -> bool:
    origin = get_origin(t)
    args = get_args(t)
    return origin is type and len(args) == 1 and isinstance(args[0], type) and issubclass(args[0], ForeignConfig)


def _structure_foreign_config_factory(
    t: object,  # noqa: ARG001
    c: cattrs.Converter,  # noqa: ARG001
) -> Callable[[object, object], type[ForeignConfig]]:
    name_to_type = {t.__name__: t for t in ForeignConfig.__descendants__()}

    def _hook(v: object, _t: object) -> type[ForeignConfig]:
        if isinstance(v, type) and issubclass(v, ForeignConfig):
            return v
        return name_to_type[cast("str", v)]

    return _hook


def _unstructure_foreign_config_factory(t: object, c: cattrs.Converter) -> Callable[[type[ForeignConfig]], str]:  # noqa: ARG001
    def _hook(v: type[ForeignConfig]) -> str:
        return v.__name__

    return _hook


def _structure_system_prompt(value: object, _type: object) -> FileSpec | str | None:
    if value is None:
        return None
    if isinstance(value, str):
        return value
    if isinstance(value, dict):
        value = cast("dict[Any, Any]", value)
        if "filename" in value:
            return FileSpec(**value)
    raise ValueError(f"Invalid system_prompt value: {value}")


def _structure_chat_template(value: object, _type: object) -> FileSpec | JSONFieldSpec | str | None:
    if value is None:
        return None
    if isinstance(value, str):
        return value
    if isinstance(value, dict):
        value = cast("dict[Any, Any]", value)  # ty bug??? Why is just `dict` != `dict[Any, Any]`?
        if "file_spec" in value and "field_name" in value:
            return JSONFieldSpec(
                file_spec=FileSpec(**value["file_spec"]),
                field_name=value["field_name"],
            )
        if "filename" in value:
            return FileSpec(**value)
    raise ValueError(f"Invalid chat_template value: {value}")


@dataclass(frozen=True)
class ModelSpec:
    _converter: ClassVar[cattrs.Converter] = cattrs.Converter()

    _converter.register_structure_hook_factory(_is_foreign_config_type, _structure_foreign_config_factory)
    _converter.register_unstructure_hook_factory(_is_foreign_config_type, _unstructure_foreign_config_factory)
    _converter.register_structure_hook(FileSpec | JSONFieldSpec | str | None, _structure_chat_template)
    _converter.register_structure_hook(FileSpec | str | None, _structure_system_prompt)

    vendor: str
    family: str
    name: str
    size: str
    repo: str
    config_type: type[ForeignConfig]
    quantization: QuantizationMode | None = None
    output_parser_regex: str | None = None
    system_role_name: str = "system"
    user_role_name: str = "user"
    assistant_role_name: str = "assistant"
    tool_role_name: str = "tool"
    weights_type: WeightsType = WeightsType.SAFETENSORS
    model_type: ModelType = ModelType.LANGUAGE_MODEL
    configs: ConfigMap = field(default=ConfigMap())
    use_cases: tuple[UseCase, ...] = tuple()
    grammar_start_tokens: tuple[str, ...] = tuple()

    @classmethod
    def from_json(cls, json_data: dict) -> "ModelSpec":
        return cls._converter.structure(json_data, cls)

    def to_json(self) -> dict:
        return self._converter.unstructure(self)


def awq_model_spec(
    model_spec: ModelSpec,
    repo: str,
    quantization: QuantizationMode = QuantizationMode.UINT4,
) -> ModelSpec:
    return ModelSpec(
        vendor=model_spec.vendor,
        family=model_spec.family,
        name=f"{model_spec.name}-AWQ",
        size=model_spec.size,
        quantization=quantization,
        repo=repo,
        config_type=model_spec.config_type,
        configs=model_spec.configs,
        weights_type=model_spec.weights_type,
        use_cases=model_spec.use_cases,
        grammar_start_tokens=model_spec.grammar_start_tokens,
    )


def build_quantized_models(model_specs: list[ModelSpec]) -> list[ModelSpec]:
    quantization_compatible_repos: list[str] = [
        "Qwen/Qwen2.5-3B-Instruct",
        "Qwen/Qwen2.5-7B-Instruct",
        "Qwen/Qwen2.5-Coder-3B-Instruct",
        "Qwen/Qwen2.5-Coder-7B-Instruct",
        "deepseek-ai/DeepSeek-R1-Distill-Qwen-1.5B",
        "HuggingFaceTB/SmolLM2-1.7B-Instruct",
        "meta-llama/Llama-3.2-3B-Instruct",
    ]

    quantized_model_specs: list[ModelSpec] = []
    for model_spec in model_specs:
        if model_spec.repo not in quantization_compatible_repos:
            continue
        quantized_repo = "trymirai/{}-AWQ".format(model_spec.repo.split("/")[-1])
        quantized_model_spec = awq_model_spec(model_spec, quantized_repo)
        quantized_model_specs.append(quantized_model_spec)
    return quantized_model_specs
